# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Functions managing TCP communications
"""

from copy import deepcopy

from qat.internal.application_context import ApplicationContext
from qat.internal.find_object import object_to_definition
from qat.test_settings import Settings


def connect(
        app_context: ApplicationContext,
        definition: dict,
        object_property: str,
        callback=None,
        timeout=Settings.wait_for_object_timeout) -> str:
    """
    Connect a signal from the application to the given callback.
    Return a unique identifier for the newly created connection
    """

    command = {}
    command['command'] = 'communication'
    command['attribute'] = 'connect'
    command['object'] = deepcopy(object_to_definition(definition))
    command['args'] = object_property

    result = app_context.send_command(command, timeout)

    if 'id' not in result:
        raise RuntimeError('Server did not return an ID for this connection')

    conn_id = result['id']
    app_context.register_callback(conn_id, callback)
    return conn_id


def disconnect(
        app_context: ApplicationContext,
        conn_id: str,
        timeout=Settings.wait_for_object_timeout) -> bool:
    """
    Disconnect a signal from its callback.
    conn_id: a unique identifier for the connection, as returned by connect()
    Return True if the signal was disconnected, False otherwise.
    """
    command = {}
    command['command'] = 'communication'
    command['attribute'] = 'disconnect'
    command['args'] = conn_id

    result = app_context.send_command(command, timeout)

    app_context.unregister_callback(conn_id)

    return 'found' in result and result['found']



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Provides a class to manage bindings
"""

import qat


class Binding():
    """
    Automatically manage a Qat connection
    """

    def __init__(
            self,
            remote_object: dict,
            remote_property: str,
            local_object,
            local_property: str):
        """
        Establish a new connection between the remote object's property
        and the local object's member
        """
        self._remote_object = remote_object
        self._remote_property = remote_property
        self._local_object = local_object
        self._local_property = local_property
        self._conn_id = None
        self.connect()


    def __del__(self):
        """
        Disconnect the binding upon destruction
        """
        try:
            self.disconnect()
        except: # pylint: disable=bare-except
            pass


    def __call__(self, *args):
        """
        Implement a functor to use this binding object as a connection callback.
        Set the local property's value upon connection update
        """
        args = list(args)
        if len(args) != 1:
            print('Invalid binding call: must have a single mandatory argument')
            return
        setattr(self._local_object, self._local_property, args[0])


    def connect(self):
        """
        Connect (or re-connect) this binding to the remote object.
        """
        self._conn_id = qat.connect(
            self._remote_object, self._remote_property, self)
        # Initialize current value
        current_value = getattr(self._remote_object, self._remote_property)
        self(current_value)


    def disconnect(self) -> bool:
        """
        Disconnect this binding. Receiver will not be updated anymore.
        """
        if self._conn_id is not None:
            result = qat.disconnect(self._conn_id)
            self._conn_id = None
            return result
        return True



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Provides the QtCustomObject class
"""

from copy import deepcopy


class QtCustomObject():
    """
    Class binding local attributes and methods to a remote Qt object
    """

    def __init__(self, attributes: dict) -> None:
        """
        Store the object's attributes
        """
        self.__dict__['_attributes'] = deepcopy(attributes)
        self.__dict__['typeId'] = None
        self.__dict__['type'] = None
        if 'QVariantType' in self.__dict__['_attributes']:
            self.__dict__['typeId'] = self.__dict__[
                '_attributes']['QVariantType']
            self.__dict__['type'] = self.__dict__[
                '_attributes']['QVariantTypeName']
            del self.__dict__['_attributes']['QVariantType']
            del self.__dict__['_attributes']['QVariantTypeName']
        # Handle nested objects
        for att in attributes:
            if isinstance(attributes[att], dict):
                self.__dict__['_attributes'][att] = QtCustomObject(attributes[att])

    def get_attribute_dict(self) -> dict:
        """
        Return the object's attributes
        """
        all_attributes = deepcopy(self.__dict__['_attributes'])
        if self.typeId is not None:
            all_attributes['QVariantType'] = self.__dict__['typeId']
        # Handle nested objects
        for att in self.__dict__['_attributes']:
            if isinstance(self.__dict__['_attributes'][att], QtCustomObject):
                all_attributes[att] = self.__dict__['_attributes'][att].get_attribute_dict()

        return all_attributes

    def __str__(self):
        """
        Custom string representation of a QtCustomObject
        """
        return str(self.__dict__['_attributes'])

    def __getattr__(self, name: str):
        """
        Get an attribute of this object.
        """
        attributes = self.__dict__['_attributes']
        if name not in attributes:
            raise AttributeError(f"Attribute {name} does not exist")

        return attributes[name]

    def __setattr__(self, name: str, value):
        """
        Set a (local) attribute of this object.
        """
        if name == 'typeId':
            self.__dict__['typeId'] = value
            return
        if name == 'type':
            self.__dict__['type'] = value
            return
        attributes = self.__dict__['_attributes']
        if name not in attributes:
            raise AttributeError(f"Attribute {name} does not exist")

        attributes[name] = value

    # Forwarding supported classic operations to underlying value

    def __len__(self):
        return self.__dict__['_attributes'].__len__()

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, QtCustomObject):
            return self.__dict__['_attributes'] == other.__dict__['_attributes']
        if isinstance(other, dict):
            return self.__dict__['_attributes'] == other
        return False

    def __getitem__(self, key):
        attributes = self.__dict__['_attributes']
        if key in attributes:
            return attributes.__getitem__(key)

        if self.type == 'QColor':
            values = [attributes['red'], attributes['green'],
                      attributes['blue'], attributes['alpha']]
            return values.__getitem__(key)
        return list(attributes.values()).__getitem__(key)

    def __setitem__(self, key, value):
        attributes = self.__dict__['_attributes']
        if key in attributes:
            return attributes.__setitem__(key, value)

        raise AttributeError(f'Setting {key} attribute is not supported')

    def __iter__(self):
        return self.__dict__['_attributes'].__iter__()

    def __contains__(self, item):
        return self.__dict__['_attributes'].__contains__(item)

    def __deepcopy__(self, memo):
        return deepcopy(self.__dict__['_attributes'])



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Sample file for Behave integration
"""

from pathlib import Path
from qat import bdd_hooks


def before_all(context):
    """
    Called before a test starts.
    """
    suite_name = Path(__file__).resolve().parent.stem
    bdd_hooks.before_all(context, suite_name)


def before_feature(context, feature):
    """
    Called before a feature starts.
    """
    bdd_hooks.before_feature(context, feature)
    context.userData = {}


def before_scenario(context, scenario):
    """
    Called before a scenario starts.
    """
    bdd_hooks.before_scenario(context, scenario)


def before_step(context, step):
    """
    Called before a step starts.
    """
    bdd_hooks.before_step(context, step)


def after_step(context, step):
    """
    Called after a step ends.
    """
    bdd_hooks.after_step(context, step)


def after_scenario(context, scenario):
    """
    Called after a scenario ends.
    """
    bdd_hooks.after_scenario(context, scenario)


def after_feature(context, feature):
    """
    Called after a feature ends.
    """
    bdd_hooks.after_feature(context, feature)


def after_all(context):
    """
    Called after a test has run.
    """
    bdd_hooks.after_all(context)



# -*- coding: utf-8 -*-

from qat import report
import qat

from scripts.object_dictionary import main_window

def test_launch():
    # Step #1: get the first registered application
    # If this step fails, run the 'qat-gui' command in the suite folder
    # to register any Qt application.
    registered_apps = qat.list_applications()
    if len(registered_apps) == 0:
        report.failed(
            'No application found',
            'Use qat-gui command or qat.register_application() to register an application'
        )
    app_name = list(registered_apps)[0]
    report.log(f"'{app_name}' is available")

    # Step #2: launch the application
    # If this step fails, verify that your application uses a supported
    # Qt version and that Qat is not blocked by any antivirus or firewall.
    app_context = qat.start_application(app_name)
    report.log(f"'{app_name}' was successful started")

    # Step #3: verify that the main window is opened
    # If this step fails, it means that there is no ApplicationWindow
    # displayed. You can edit its definition in scripts/object_dictionary.py.
    window = qat.wait_for_object_exists(main_window)

    report.verify(
        window.visible,
        "Main Window is opened",
        f"Property 'visible' is {window.visible}"
    )

    # Step #4: close the application
    qat.close_application(app_context)


if __name__ == "__main__":

    # Before running any test, you need to register the tested application.
    # You can do so by starting the 'qat-gui' command in the current directory
    # or by calling the qat.register_application() function:
    # qat.register_application(app_name, args)
    # Please refer to API documentation for details.

    # Create a test report
    report.start_report('demo', './report.xml')

    try:
        test_launch()
    finally:
        # This call is mandatory when the test ends, without it,
        # the report writer would still be running and the execution
        # would hang. It is recommended to use the 'teardown' feature of your
        # test framework (e.g. pytest) to guarantee this call is made.
        report.stop_report()



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Local definition of Qat constants
"""

from dataclasses import dataclass
import json
from qat.internal import global_constants

@dataclass
class Settings:
    """
    Data class holding local definition of constants.
    Values can be changed from a testSettings.json file.
    """
    wait_for_object_timeout = global_constants.WAIT_FOR_OBJECT_TIMEOUT
    wait_for_app_start_timeout = global_constants.WAIT_FOR_APP_START
    wait_for_app_stop_timeout = global_constants.WAIT_FOR_APP_STOP
    long_operation_timeout = global_constants.LONG_OPERATION_TIMEOUT

    screenshot_on_fail = global_constants.SCREENSHOT_ON_FAIL
    continue_after_fail = global_constants.CONTINUE_AFTER_FAIL
    lock_ui = global_constants.LOCK_UI

    loaded = False


def load_test_settings():
    """
    Parse the testSettings.json file to initialize constant values
    """
    if Settings.loaded:
        return
    try:
        with open('testSettings.json', 'rt', encoding='utf-8') as file:
            settings = json.load(file)
            if 'waitForObjectTimeout' in settings:
                Settings.wait_for_object_timeout = settings['waitForObjectTimeout']
            if 'waitForAppStartTimeout' in settings:
                Settings.wait_for_app_start_timeout = settings['waitForAppStartTimeout']
            if 'waitForAppStopTimeout' in settings:
                Settings.wait_for_app_stop_timeout = settings['waitForAppStopTimeout']
            if 'longOperationTimeout' in settings:
                Settings.long_operation_timeout = settings['longOperationTimeout']
            if 'screenshotOnFail' in settings:
                Settings.screenshot_on_fail = settings['screenshotOnFail']
            if 'continueAfterFail' in settings:
                Settings.continue_after_fail = settings['continueAfterFail']
            if 'lockUI' in settings:
                Settings.lock_ui = settings['lockUI']
            Settings.loaded = True
    except: # pylint: disable=bare-except
        print("Could not load testSettings.json, using default values.")



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Functions related to timing and synchronization
"""

import time

from collections.abc import Callable

from qat.test_settings import Settings
from qat.internal.application_context import ApplicationContext
from qat.internal.find_object import wait_for_object_exists


def wait_for_property(
        app_context: ApplicationContext,
        definition: dict,
        property_name: str,
        reference,
        comparator: Callable,
        timeout = Settings.wait_for_object_timeout,
        check = False) -> bool:
    """
    Wait for the given property to reach the given condition.
    property_name can be a nested property (using dot notation, e.g. 'color.name')
    """
    start_time = round(1000 * time.time())

    last_error = None
    nested_properties = property_name.split('.')
    while (round(1000 * time.time()) - start_time) < timeout:
        try:
            last_error = None
            qtobject = wait_for_object_exists(app_context, definition, timeout)
            current_object = qtobject
            for prop in nested_properties:
                value = getattr(current_object, prop)
                current_object = value
            if comparator(value, reference):
                return True
            raise ValueError(f'Property "{property_name}" is "{value}"')
        except (LookupError, ValueError, AttributeError) as error:
            last_error = error
            time.sleep(0.1)

    if not check:
        if last_error is not None:
            print(f'ERROR: {last_error}')
        return False

    if last_error is not None:
        raise last_error
    raise TimeoutError("Property did not reach the given condition before timeout")



# -*- coding: utf-8 -*-

"""
Then steps implementation
"""

from behave import Then

from qat import report
import qat

from scripts.object_dictionary import bdd_mapping

@Then("The {name} is opened")
def step(_, name):
    object_definition = bdd_mapping[name]

    window = qat.wait_for_object_exists(object_definition)

    report.verify(
        window.visible,
        f"{name} is opened",
        f"Property 'visible' is {window.visible}"
    )




# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Public Qat API
"""

# Some imports are only to make types accessible with qat.Type without using 'internal' namespace
# pylint: disable=unused-import

# Mouse events require many arguments
# pylint: disable = too-many-arguments


from dataclasses import dataclass
from pathlib import Path
from typing import Optional
import json
import os
import time

from qat.internal import app_launcher
from qat.internal.application_context import ApplicationContext
from qat.internal.qt_custom_object import QtCustomObject
from qat.internal.qt_object import QtObject
import qat.internal.communication_operations as comm
import qat.internal.debug_operations as debug
import qat.internal.find_object as find
import qat.internal.mouse_operations as mouse
import qat.internal.gesture_operations as gesture
import qat.internal.keyboard_operations as keyboard
import qat.internal.screenshot_operations as screenshot
import qat.internal.synchronization as sync
import qat.internal.touch_operations as touch

from qat.internal.mouse_operations import Button, Modifier
from qat.internal.xml_report import XmlReport
from qat.internal.binding import Binding

from qat.test_settings import Settings

@dataclass
class Globals:
    """Holds the global state of the API."""
    current_report = None
    current_app_context = None


###################################################################################
# Application management and configuration
###################################################################################


def register_application(name: str, path: str, args='', shared: Optional[bool] = False) -> None:
    """
    Add the given application to the configuration file.

    Args:
      name: Must be a unique name for this application. It will be used when calling :func:`start_application` for example.
      path: The absolute path to the application executable.
      args (optional): The default arguments used when launching the application (e.g. when using Qat-Spy). 
        They can be overridden when calling :func:`start_application`.
      shared (optional): True to use the shared (global) file, False to use the local one. 
        If None (default), the local file will be used if it exists otherwise the shared one will be used.
    """
    config_file = get_config_file(shared)
    applications = {}
    try:
        with open(config_file, 'r', encoding='utf-8') as file:
            applications = json.load(file)
    except: # pylint: disable=bare-except
        print("Invalid configuration file. Content will be overwritten.")
    applications[name] = {
        "path": path,
        "args": args
    }
    if not config_file.is_file():
        os.makedirs(config_file.parent, exist_ok=True)
        with open(config_file, "w", encoding="utf-8") as file:
            file.write("{}")
            file.close()
    with open(config_file, 'w', encoding='utf-8') as file:
        json.dump(applications, file, indent=3)


def unregister_application(name: str, shared: Optional[bool] = False) -> None:
    """
    Remove the given application from the configuration file.

    Args:
      name: The unique name of the application, must be the same as the one given to :func:`register_application`.
      shared: True to use the shared (global) file, False to use the local one.
        If None, the local file will be used if it exists otherwise the shared one will be used.
    """
    config_file = get_config_file(shared)
    applications = {}
    if os.path.exists(config_file):
        with open(config_file, 'r', encoding='utf-8') as file:
            applications = json.load(file)

    if name in applications:
        del applications[name]
        if len(applications) > 0:
            with open(config_file, 'w', encoding='utf-8') as file:
                json.dump(applications, file, indent=3)
        else:
            os.remove(config_file)


def list_applications() -> dict:
    """
    List all registered applications (shared and local).
    
    The contents of both local and shared configurations is merged.
    Values from the local configuration take precedence: if an application is 
    registered in both configurations with different paths or arguments, 
    the values from the local configuration will be returned.
    """
    return app_launcher.list_applications()


def get_config_file(shared: Optional[bool] = None) -> Path:
    """
    Return the absolute path to the current configuration file (applications.json).

    Args:
      shared: True to use the shared (global) file, False to use the local one.
    
    Returns:
      The absolute path to the current configuration file.

      If shared is None and if a configuration file exists in the current working directory, 
      it will be returned. Otherwise, the global one from the user directory (e.g. ~/.qat) 
      will be returned if it exists. 
      Finally, if no file exists, the path to a default file in the current working directory will be returned.
    """
    return app_launcher.get_config_file(shared)


def get_application_path(name: str) -> str:
    """
    Return the path of the given application from the configuration file.

    Args:
      name: The unique name of the application, must be the same as the one given to :func:`register_application`.
    """
    applications = list_applications()
    return applications[name]['path']


###################################################################################
# Application life cycle
###################################################################################

def start_application(app_name: str, args: str = None, detached: bool = False) -> ApplicationContext:
    """
    Start a registered application.

    Start the given application, inject the server library (except if detached is True)
    and return the corresponding application context.

    Args:
      app_name: The unique name of the application, must be the same as the one given to :func:`register_application`.
      args: A string containing arguments to be passed to the application. 
        If None (default), arguments will be read from the current configuration file.
      detached: If True, the application will start without initializing Qat. 
        This is useful when testing cases where the application returns early (e.g. in case of errors such as invalid arguments).
    
    Returns:
      An ApplicationContext that uniquely identifies the started application instance.
      Even when 'detached' argument is True, this ApplicationContext can still be used to retrieve 
      the corresponding exit code and process ID.
    """
    Globals.current_app_context = None
    Globals.current_app_context = app_launcher.start_application(app_name, args, detached)
    return Globals.current_app_context


def attach_to_application(name_or_pid) -> ApplicationContext:
    """
    Attach to the given application by name, file name or process ID.

    If a name is given, it must correspond to a registered application.
    On linux, the application must have been launched by Qat (by calling :func:`start_application`) for this function to work.

    Args:
      name_or_pid: Can be the name of a registered application, the name of a (running) executable file or a process ID.

    Returns:
      An ApplicationContext that uniquely identifies the attached application instance.
    
    Raises:
      ProcessLookupError: The application process could not be found or failed to connect.
    """
    Globals.current_app_context = app_launcher.attach_to(name_or_pid)
    return Globals.current_app_context


def current_application_context() -> ApplicationContext:
    """Return the current application context."""
    return Globals.current_app_context


def set_current_application_context(app_context: ApplicationContext):
    """
    Change the current application context.

    All subsequent API calls will use this context until this function is called again.
    """
    Globals.current_app_context = app_context


def close_application(app_context = None) -> int:
    """
    Terminate the application associated to the given context and returns its exit code.

    Args:
      app_context: The context of the application to be closed. If None, the current context will be used.

    Returns:
      The exit code returned by the closed application.
      This usually corresponds to SIGKILL since the application is killed when calling this function.

    Raises:
      ProcessLookupError: No process were found for the given application context.
    """
    if app_context is None:
        app_context = current_application_context()
    if app_context:
        app_context.kill()
        exit_code = app_context.get_exit_code()
        Globals.current_app_context = None
        return exit_code
    raise ProcessLookupError("Cannot close application: process does not exist")


###################################################################################
# Application windows
###################################################################################

def lock_application():
    """Lock application by filtering external/user events."""
    debug.lock_application(current_application_context())


def unlock_application():
    """Unlock application by allowing external/user events."""
    debug.unlock_application(current_application_context())


def list_top_windows() -> list:
    """Return all the top windows of the application."""
    return find.list_top_windows(Globals.current_app_context)


###################################################################################
# Accessing objects and widgets
###################################################################################

def find_all_objects(definition: dict) -> list:
    """
    Return all objects matching the given definition.

    Args:
      definition: A dictionary of property names and values.
    
    Returns:
      A (potentially empty) list of QtObjects corresponding to the given properties.
    """
    return find.find_all_objects(Globals.current_app_context, definition)


def wait_for_object_exists(definition: dict, timeout=None) -> QtObject:
    """
    Wait for the given object to exist in the tested application.

    Args:
      definition: A dictionary of property names and values identifying a unique object.
    
    Raises:
      LookupError: No object was found or multiple objects match the given definition.
    """
    if timeout is None:
        timeout = Settings.wait_for_object_timeout
    return find.wait_for_object_exists(Globals.current_app_context, definition, timeout)


def wait_for_object(definition: dict, timeout=None) -> QtObject:
    """
    Wait for the given object to be accessible in the application.

    This function is similar to t:func:`wait_for_object_exists` but also 
    waits for the object to be accessible (i.e. visible and enabled).

    Args:
      definition: A dictionary of property names and values identifying a unique object.
      timeout: Maximum time in milliseconds that this function will wait before raising a LookupError.
        If None (default) the timeout value will be read from the testSettings.json file.
    
    Raises:
      LookupError: No object was found before the given timeout or multiple objects match the given definition.
    """
    if timeout is None:
        timeout = Settings.wait_for_object_timeout
    return find.wait_for_object(Globals.current_app_context, definition, timeout)


def wait_for_object_missing(definition: dict, timeout=None):
    """
    Wait for the given object to be deleted from the application.

    Args:
      definition: A dictionary of property names and values identifying a unique object.
      timeout: Maximum time in milliseconds that this function will wait before raising a LookupError.
        If None (default) the timeout value will be read from the testSettings.json file.

    Raises:
      TimeoutError: The given object still exists after the given timeout (in milliseconds) has been reached.
    """
    if timeout is None:
        timeout = Settings.wait_for_object_timeout
    find.wait_for_object_missing(Globals.current_app_context, definition, timeout)


###################################################################################
# Accessing properties
###################################################################################

def wait_for_property_value(
    definition: dict,
    property_name: str,
    new_value,
    comparator = None,
    check = False,
    timeout = None) -> bool:
    """
    Wait for the given object's property to reach the given value.

    Args:
      definition: A QtObject or an object definition.
      property_name: The name of the property.
      new_value: The value to reach.
      comparator: Callable used to compare property values. == is used by default.
      check: If True, raises an exception in case of failure. False by default.
      timeout: If the new_value is not reached after this timeout (in milliseconds), returns False.

    Returns:
      True if the value was reached, False otherwise.
    """
    if timeout is None:
        timeout = Settings.wait_for_object_timeout

    def default_compare(value, reference):
        return value == reference

    return sync.wait_for_property(
        Globals.current_app_context,
        definition,
        property_name,
        new_value,
        comparator or default_compare,
        timeout,
        check)


def wait_for_property_change(
    definition: dict,
    property_name: str,
    old_value,
    comparator = None,
    check = False,
    timeout = None) -> bool:
    """
    Wait for the given object's property to change its value.

    Args:
      definition: A QtObject or an object definition.
      property_name: The name of the property.
      old_value: The original value.
      comparator: Callable used to compare property values. == is used by default.
      check: If True, raises an exception in case of failure. False by default.
      timeout: If the new_value has not changed after this timeout (in milliseconds), returns False.

    Returns:
      True if the value was changed, False otherwise.
    """
    if timeout is None:
        timeout = Settings.wait_for_object_timeout

    def default_compare(value, reference):
        return value != reference

    def reverse_comparator(value, reference):
        return not comparator(value, reference)

    return sync.wait_for_property(
        Globals.current_app_context,
        definition,
        property_name,
        old_value,
        reverse_comparator if comparator else default_compare,
        timeout,
        check)


def wait_for(condition, timeout=None) -> bool:
    """
    Wait for the given condition to be reached.

    Args:
      condition: Any expression, function or lambda returning a boolean.
      timeout: If the condition is not reached after this timeout (in milliseconds), returns False.

    Returns:
      True if the condition was reached before timeout, False otherwise
    """
    if timeout is None:
        timeout = Settings.wait_for_object_timeout
    start_time = round(1000 * time.time())
    reached = False
    while not reached and (round(1000 * time.time()) - start_time) < timeout:
        reached = condition()
        if reached:
            break
        time.sleep(0.2)

    return reached


###################################################################################
# Mouse interactions
###################################################################################

def mouse_press(
        definition: dict,
        x=None,
        y=None,
        modifier=Modifier.NONE,
        button=Button.LEFT,
        check=False):
    """
    Press a mouse button on the given widget using the given parameters.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: The coordinates of the event, relative to the object. If not given, event will occur at the center of the object.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
      button: Optional mouse button. Must be one of the constants in qat.Button. Default is LEFT.
      check: Optional check. When True, this function will raise an exception if no widget accepted the event. Default is False.
    """
    mouse.click(Globals.current_app_context, definition,
                "press", x, y, modifier, button, check)


def mouse_release(
        definition: dict,
        x=None,
        y=None,
        modifier=Modifier.NONE,
        button=Button.LEFT,
        check=False):
    """
    Release a mouse button on the given widget using the given parameters.
    
    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: The coordinates of the event, relative to the object. If not given, event will occur at the center of the object.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
      button: Optional mouse button. Must be one of the constants in qat.Button. Default is LEFT.
      check: Optional check. When True, this function will raise an exception if no widget accepted the event. Default is False.
    """
    mouse.click(Globals.current_app_context, definition,
                "release", x, y, modifier, button, check)


def mouse_click(
        definition: dict,
        x=None,
        y=None,
        modifier=Modifier.NONE,
        button=Button.LEFT,
        check=False):
    """
    Click on the given widget using the given parameters.
        
    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: The coordinates of the event, relative to the object. If not given, event will occur at the center of the object.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
      button: Optional mouse button. Must be one of the constants in qat.Button. Default is LEFT.
      check: Optional check. When True, this function will raise an exception if no widget accepted the event. Default is False.
        """
    mouse.click(Globals.current_app_context, definition,
                "click", x, y, modifier, button, check)


def double_click(
        definition: dict,
        x=None,
        y=None,
        modifier=Modifier.NONE,
        button=Button.LEFT,
        check=False):
    """
    Double-click on the given widget using the given parameters.
        
    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: The coordinates of the event, relative to the object. If not given, event will occur at the center of the object.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
      button: Optional mouse button. Must be one of the constants in qat.Button. Default is LEFT.
      check: Optional check. When True, this function will raise an exception if no widget accepted the event. Default is False.
    """
    mouse.click(Globals.current_app_context, definition,
                "double-click", x, y, modifier, button, check)


def mouse_move(
        definition: dict,
        x=None,
        y=None,
        modifier=Modifier.NONE,
        button=Button.LEFT):
    """
    Move the mouse using the given parameters.

    If :func:`mousePress` was called before, this will act as drag operation.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: the coordinates of the event, relative to the object. If not given, event will occur at the center of the object.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
      button: Optional mouse button. Must be one of the constants in qat.Button. Default is LEFT.
    """
    mouse.click(Globals.current_app_context, definition,
                "move", x, y, modifier, button)


def mouse_drag(
        definition: dict,
        x=None,
        y=None,
        dx=None,
        dy=None,
        modifier=Modifier.NONE,
        button=Button.LEFT,
        check=False):
    """
    Drag the mouse using the given parameters.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: The coordinates of the event, relative to the object. If not given, event will occur at the center of the object.
      dx, dy: the number of pixels to move the mouse by, in each direction.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
      button: Optional mouse button. Must be one of the constants in qat.Button. Default is LEFT.
      check: Optional check. When True this function will raise an exception if no widget accepted the event. Default is False.
    """
    mouse.drag(Globals.current_app_context, definition, 'drag', x,
               y, dx, dy, modifier, button, check=check)


def mouse_wheel(
        definition: dict,
        x=None,
        y=None,
        x_degrees=0,
        y_degrees=0,
        modifier=Modifier.NONE,
        check=False):
    """
    Scroll the mouse by x_degrees, y_degrees at x,y position.

    Default degree increment should be 15 to represent one physical rotation increment.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: the coordinates of the event, relative to the object. If not given, event will occur at the center of the object.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
      button: Optional mouse button. Must be one of the constants in qat.Button. Default is LEFT.
      check: Optional check. When True this function will raise an exception if no widget accepted the event. Default is False.
    """
    mouse.drag(Globals.current_app_context, definition, 'scroll', x, y, 8 * x_degrees,
               8 * y_degrees, modifier=modifier, button=Button.MIDDLE, check=check)


###################################################################################
# Touch screen/pad interactions
###################################################################################

def touch_press(
        definition: dict,
        x=None,
        y=None,
        modifier=Modifier.NONE):
    """
    Press fingers on the given widget using the given parameters.

    Press one or more fingers on the definition object at local coordinates x, y while holding the modifier key.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: the coordinates of each finger, relative to the object. If not given, event will occur at the center of the object. 
        Each argument can be a single value (for single touch point) or an array (for multiple touch points). 
        When using arrays, x and y must be of the same size.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
    """
    touch.tap(Globals.current_app_context, definition,
                "press", x, y, modifier)


def touch_release(
        definition: dict,
        x=None,
        y=None,
        modifier=Modifier.NONE):
    """
    Release fingers on the given widget using the given parameters.

    Release one or more fingers on the definition object at local coordinates x, y while holding the modifier key.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: the coordinates of each finger, relative to the object. If not given, event will occur at the center of the object. 
        Each argument can be a single value (for single touch point) or an array (for multiple touch points). 
        When using arrays, x and y must be of the same size.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
    """
    touch.tap(Globals.current_app_context, definition,
                "release", x, y, modifier)


def touch_tap(
        definition: dict,
        x=None,
        y=None,
        modifier=Modifier.NONE):
    """
    Tap on the given widget using the given parameters.

    Tap one or more fingers on the definition object at local coordinates x, y while holding the modifier
    Note that sending two touch_tap events quickly will not generate a double-tap event. Use "func:`double_click` instead.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: the coordinates of each finger, relative to the object. If not given, event will occur at the center of the object. 
        Each argument can be a single value (for single touch point) or an array (for multiple touch points). 
        When using arrays, x and y must be of the same size.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
    """
    touch.tap(Globals.current_app_context, definition,
                "tap", x, y, modifier)


def touch_move(
        definition: dict,
        x=None,
        y=None,
        modifier=Modifier.NONE):
    """
    Move fingers on the given widget using the given parameters.

    Move one or more fingers on the definition object at local coordinates x, y while holding the modifier key.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: the coordinates of each finger, relative to the object. If not given, event will occur at the center of the object. 
        Each argument can be a single value (for single touch point) or an array (for multiple touch points). 
        When using arrays, x and y must be of the same size.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
    """
    touch.tap(Globals.current_app_context, definition,
                "move", x, y, modifier)


def touch_drag(
        definition: dict,
        x=None,
        y=None,
        dx=0,
        dy=0,
        modifier=Modifier.NONE):
    """
    Drag fingers on the given widget using the given parameters.

    Press and drag one or more fingers on the definition object starting at local coordinates x, y while holding the modifier key.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      x, y: the coordinates of each finger, relative to the object. If not given, event will occur at the center of the object. 
        Each argument can be a single value (for single touch point) or an array (for multiple touch points). 
        When using arrays, x and y must be of the same size.
      dx, dy: the number of pixels to move each finger, relative to the object. 
        Each argument can be a single value (to apply the same movement to all touch points) or 
        an array (to have different movements for each touch point). 
        When using arrays, dx and dy must be of the same size.
      modifier: Optional keyboard modifier. Must be one of the constants in qat.Modifier. Default is NONE.
    """
    touch.drag(Globals.current_app_context, definition, 'drag', x,
               y, dx, dy, modifier)


###################################################################################
# Gestures
###################################################################################

def flick(
        definition: dict,
        dx=0,
        dy=0):
    """
    Move the given Flickable by the given horizontal and vertical distances in pixels.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      dx: the horizontal distance in pixels.
      dy: the vertical distance in pixels.
    """
    gesture.flick(Globals.current_app_context, definition, dx, dy)


def pinch(
        definition: dict,
        rotation = 0.,
        translation = None,
        scale = 1.0):
    """
    Generate a pinch event (zoom, rotation, pan).

    All three optional parameters can be combined to define a complete pinch gesture.
    The distance between the two fingers during the gesture will be automatically determined based on the target widget’s size.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      rotation (optional): Total angle of rotation to apply. Value is in degrees and in clockwise direction.
      translation (optional): Global translation of the pinch. This corresponds to the translation 
        of the central point between the two fingers.
      scale (optional): Variation of the distance between the two fingers, typically representing a zoom factor. 
        For example, if scale is set to 2.0, the distance between the fingers will double between the beginning and the end of the gesture.
    """
    gesture.pinch(
        Globals.current_app_context,
        definition,
        rotation,
        translation,
        scale
    )


def native_pinch(
        definition: dict,
        angle: int = None,
        scale: float = None,
        check: bool = False):
    """
    Generate a native pinch event (zoom and/or rotation).

    Native gesture events are high-level events generated by the operating system, usually from a sequence of trackpad events.
    This function works with QML/QtQuick widgets only.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      angle (optional): Total angle of rotation to apply. Value is in degrees and in clockwise direction.
      scale (optional): Variation of the distance between the two fingers, typically representing a zoom factor. 
        For example, if scale is set to 2.0, the distance between the fingers will double between the beginning and the end of the gesture.
      check (optional): If True, raises an exception in case of failure (i.e when the event is not handled by any widget). False by default.
    """
    gesture.native_pinch(
        Globals.current_app_context,
        definition,
        angle,
        scale,
        check)


###################################################################################
# Keyboard interactions
###################################################################################

def type_in(
        definition: dict,
        text: str):
    """
    Type the given text in the given object.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      text: Any string. Can also use the following special keys: 
        <Backspace>, <Delete>, <Enter>, <Escape>, <Return>, <Tab>, <Control>, <Shift>, <Alt>
    """
    keyboard.type_in(Globals.current_app_context, definition, text)


def shortcut(
        definition: dict,
        key_combination: str):
    """
    Trigger the given shortcut on the given object.

    Shortcut string must follow the Qt syntax, e.g:
    'Ctrl+Z, Alt+O, Alt+Shift+R, ...'

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      key_combination: The key combination to trigger.
    """
    keyboard.shortcut(Globals.current_app_context, definition, key_combination)


def press_key(
        definition: dict,
        key: str):
    """
    Press the given key on the given object.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      text: Any character. Can also use the following special keys: 
        <Backspace>, <Delete>, <Enter>, <Escape>, <Return>, <Tab>, <Control>, <Shift>, <Alt>
    """
    keyboard.press_key(Globals.current_app_context, definition, key)


def release_key(
        definition: dict,
        key: str):
    """
    Release the given key on the given object.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      text: Any character. Can also use the following special keys: 
        <Backspace>, <Delete>, <Enter>, <Escape>, <Return>, <Tab>, <Control>, <Shift>, <Alt>
    """
    keyboard.release_key(Globals.current_app_context, definition, key)



###################################################################################
# Screenshots
###################################################################################

def take_screenshot(path=None):
    """
    Take a screenshot of each current top-level window.
     
    Screenshots will be saved to the given path. If no path is provided, screenshots will be
    saved to the 'screenshots' subfolder of the current directory.
    The file name will be qat-screenshotN.png, where N is the index of the window.

    Args:
      path: Path to save the screenshots. Extension will determine the format of the image file (png, jpg, bmp, ...).
    """
    screenshot.take_screenshot(Globals.current_app_context, path)


def grab_screenshot(definition, delay=0, timeout=None):
    """
    Take a screenshot of the given widget after an optional delay in ms.

    Args:
      definition: An object definition or an Object returned by :func:`wait_for_object*`.
      delay: Wait delay milliseconds before taking the screenshot. Default is 0: no delay.
      timeout: Number of milliseconds to wait for the screenshot to be available once taken. If timeout is reached, an exception is raised.

    Returns:
      An Image object which provides the following functions:
        def getPixel(x: int, y: int) -> int
        def getPixelRGBA(x: int, y: int) -> Color
        def save(file_name: str)

    Raises:
      LookupError: The given object could not be found.
      RuntimeError: The image could not be generated.
    """
    if timeout is None:
        timeout = Settings.wait_for_object_timeout
    return screenshot.grab_screenshot(Globals.current_app_context, definition, delay, timeout)



###################################################################################
# Picking
###################################################################################

def activate_picker():
    """Activate the object picker"""
    debug.activate_picker(Globals.current_app_context)


def deactivate_picker():
    """Deactivate the object picker"""
    debug.deactivate_picker(Globals.current_app_context)



###################################################################################
# Connections and bindings
###################################################################################

def connect(
        object_def: dict,
        object_property: str,
        callback) -> str:
    """
    Connect a signal from the application to the given callback.

    If 'object_property' is a signal name, the given callback will be called without argument.
    If 'object_property' is a Qt property name, the given callback will be called with
    one argument containing the new value of the property.

    Args:
      object_def: An object definition or an Object returned by :func:`wait_for_object*`.
      property: The name of the property or signal to connect to.
      callback: A Python Callable (e.g. function or lambda).

    Returns:
      A unique identifier for the newly created connection.
    """
    return comm.connect(Globals.current_app_context, object_def, object_property, callback)


def disconnect(conn_id: str):
    """
    Disconnect a signal from its callback.

    Args:
      conn_id: A connection identifier, as returned by :fund:`connect`.
    """
    return comm.disconnect(Globals.current_app_context, conn_id)


def bind(
        remote_object: dict,
        remote_property: str,
        local_object,
        local_property: str) -> Binding:
    """
    Automatically establish a connection between the given object's property and the given receiver.

    Note: this is equivalent to create a Binding object with the same arguments.

    Args:
      remote_object: An object definition or an Object returned by :func:`wait_for_object*`.
      remote_property: The name of the property or signal to connect to.
      local_object: Any Python object.
      local_attribute: The name of the Python attribute to be connected. 
        Must be an attribute of local_object and be of a compatible type.
    
    Returns:
      A Binding object that can be used to manage the connection with the following functions:
        def connect(): Connect (or re-connect) this binding to the remote object.
        def disconnect(): Disconnect this binding. Receiver will not be updated anymore.
    """
    return Binding(remote_object, remote_property, local_object, local_property)



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Functions related to keyboard
"""

from qat.internal.application_context import ApplicationContext
from qat.internal import find_object


def type_in(
        app_context: ApplicationContext,
        definition: dict,
        text: str):
    """
    Type the given text in the given object.
    """
    definition = find_object.object_to_definition(definition)

    command = {}
    command['command'] = 'keyboard'
    command['object'] = definition
    command['attribute'] = 'type'
    command['args'] = text

    app_context.send_command(command)


def shortcut(
        app_context: ApplicationContext,
        definition: dict,
        key_combination: str):
    """
    Trigger the given shortcut on the given object.
    """

    definition = find_object.object_to_definition(definition)

    command = {}
    command['command'] = 'keyboard'
    command['object'] = definition
    command['attribute'] = 'shortcut'
    command['args'] = key_combination

    app_context.send_command(command)


def press_key(
        app_context: ApplicationContext,
        definition: dict,
        key: str):
    """
    Press the given key on the given object.
    """
    definition = find_object.object_to_definition(definition)

    command = {}
    command['command'] = 'keyboard'
    command['object'] = definition
    command['attribute'] = 'press'
    command['args'] = key

    app_context.send_command(command)


def release_key(
        app_context: ApplicationContext,
        definition: dict,
        key: str):
    """
    Release the given key on the given object.
    """
    definition = find_object.object_to_definition(definition)

    command = {}
    command['command'] = 'keyboard'
    command['object'] = definition
    command['attribute'] = 'release'
    command['args'] = key

    app_context.send_command(command)



# -*- coding: utf-8 -*-
# (c) Copyright 2024, Qat’s Authors

"""
Provides utilities to help work with GUI components.
"""

import tkinter as tk
from tktooltip import ToolTip


def create_button_tooltip(
    widget : tk.Widget,
    delay: float = 0.0):
    """
    Create a tooltip to all button under the widget.
    Assign button text to the tooltip.  

    Parameters
    ----------
    widget : tk.Widget
        The parent widget where to create the tooltip
    delay : float, optional
        Delay in seconds before the ToolTip appears, by default 0.0
    """
    for children in widget.winfo_children():
        if isinstance(children, tk.Button):
            ToolTip(children, msg=children['text'], delay=delay)



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Qat Spy application
"""

from pathlib import Path
import os
import tkinter as tk
try:
    import sv_ttk
except ImportError:
    print('SunValley theme unavailable: using default ttk theme')

from qat.gui.application_manager import ApplicationManager


def generate_root_window() -> tk.Tk:
    """
    Create the main window
    """
    current_path = Path(os.path.dirname(__file__)).resolve().absolute()
    image_dir = current_path / 'images'
    root = tk.Tk()
    root.iconphoto(True, tk.PhotoImage(file=image_dir / "icon.png"))
    root.title('Qat Gui')
    root.withdraw()

    return root


def open_gui():
    """
    Open the main window
    """

    window = generate_root_window()
    ApplicationManager(window)

    try:
        sv_ttk.set_theme('light')
    except NameError:
        pass
    window.mainloop()


if __name__ == "__main__":

    open_gui()



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Implementation of the XML report writer
"""

from pathlib import Path
from xml.sax.saxutils import escape

import io
import os
import time
import traceback as tb

import threading
import queue

TIME_FORMAT = '%Y-%m-%dT%H:%M:%S'


class AsyncWriter(threading.Thread):
    """
    Write the report asynchronously
    """
    def __init__(self, report_filename: Path):
        threading.Thread.__init__(self)
        self._writer_queue = queue.Queue(100)
        # File must remain open
        # pylint: disable = consider-using-with
        self._report = io.open(report_filename, "w", encoding='utf8')
        self._stopped = False


    def __del__(self):
        self.close()


    def close(self):
        """
        Stop writing
        """
        self._stopped = True
        if self._report is not None:
            while not self._writer_queue.empty():
                time.sleep(0.1)
            self._report.flush()
            self._report.close()
            self._report = None


    def run(self):
        """
        Start writing
        """
        if self._report is None:
            return
        while not self._stopped or not self._writer_queue.empty():
            self._report.write(self._writer_queue.get())


    def write(self, data: str):
        """
        Write the given string asynchronously
        """
        if not self._stopped:
            self._writer_queue.put(data)


class XmlReport():
    """
    Public interface of the XML report
    """
    def __init__(self, suite_name: str, report_filename: Path) -> None:
        report_filename = Path(report_filename)
        self._filename = str(report_filename)
        try:
            os.makedirs(report_filename.parent, exist_ok=True)
            if report_filename.exists():
                os.remove(report_filename)

            self._report = AsyncWriter(report_filename)
            self._report.start()

            timestamp = time.strftime(TIME_FORMAT, time.localtime())
            header = []
            header.append('<?xml version="1.0" encoding="UTF-8"?>')
            header.append('<TestReport version="1.0" >')
            header.append(' <test type="testsuite">')
            header.append(f'  <prolog time="{timestamp}">')
            header.append(f'   <name>{suite_name}</name>')
            header.append('  </prolog>')
            self._report.write("\n".join(header) + "\n")
        except:
            print("Unable to create report")
            raise


    def __del__(self):
        self.end_report()


    def get_folder(self) -> Path:
        """
        Return the path to the report folder
        """
        return Path(self._filename).parent.resolve()


    def end_report(self):
        """
        Write final tags and close the report
        """
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        footer = []
        footer.append(f'  <epilog time="{timestamp}"/>')
        footer.append(' </test>')
        footer.append('</TestReport>')
        self._report.write("\n".join(footer) + "\n")
        self._report.close()
        self._report.join()


    def start_test_case(self, name):
        """
        Write tags to start a test case
        """
        indent = 2 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent +  '<test type="testcase">')
        lines.append(indent + f' <prolog time="{timestamp}">')
        lines.append(indent + f'  <name>{escape(name)}</name>')
        lines.append(indent +  ' </prolog>')
        self._report.write("\n".join(lines) + "\n")


    def end_test_case(self):
        """
        Write tags to finish a test case
        """
        indent = 2 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent + f' <epilog time="{timestamp}"/>')
        lines.append(indent +  '</test>')
        self._report.write("\n".join(lines) + "\n")


    def start_feature(self, name, description):
        """
        Write tags to start a feature
        """
        indent = 3 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent +  '<test type="feature">')
        lines.append(indent + f' <prolog time="{timestamp}">')
        lines.append(indent + f'  <name>{escape(name)}</name>')
        lines.append(
            indent + f'  <description>{escape(description)}</description>')
        lines.append(indent + ' </prolog>')
        self._report.write("\n".join(lines) + "\n")


    def end_feature(self):
        """
        Write tags to finish a feature
        """
        indent = 3 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent + f' <epilog time="{timestamp}"/>')
        lines.append(indent +  '</test>')
        self._report.write("\n".join(lines) + "\n")


    def start_scenario(self, name, description):
        """
        Write tags to start a scenario
        """
        indent = 4 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent +  '<test type="scenario">')
        lines.append(indent + f' <prolog time="{timestamp}">')
        lines.append(indent + f'  <name>{escape(name)}</name>')
        lines.append(
            indent + f'  <description>{escape(description)}</description>')
        lines.append(indent + ' </prolog>')
        self._report.write("\n".join(lines) + "\n")


    def start_example(self, headings, values):
        """
        Add example to current scenario
        """
        indent = 5 * " "
        lines = []
        lines.append(indent +  '<example>')
        for name, value in zip(headings, values):
            lines.append(indent +  f' <{name}>{value}</{name}>')
        lines.append(indent +  '</example>')
        self._report.write("\n".join(lines) + "\n")


    def end_scenario(self):
        """
        Write tags to finish a scenario
        """
        indent = 4 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent + f' <epilog time="{timestamp}"/>')
        lines.append(indent +  '</test>')
        self._report.write("\n".join(lines) + "\n")


    def start_step(self, name):
        """
        Write tags to start a step
        """
        indent = 5 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent +  '<test type="step">')
        lines.append(indent + f' <prolog time="{timestamp}">')
        lines.append(indent + f'  <name>{escape(name)}</name>')
        lines.append(indent +  ' </prolog>')
        self._report.write("\n".join(lines) + "\n")


    def end_step(self):
        """
        Write tags to finish a step
        """
        indent = 5 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent + f' <epilog time="{timestamp}"/>')
        lines.append(indent +  '</test>')
        self._report.write("\n".join(lines) + "\n")


    def log(self, text: str, log_type="LOG"):
        """
        Log the given message
        """
        indent = 6 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent + f'<message time="{timestamp}" type="{log_type}">')
        lines.append(indent + f'  <text>{escape(text)}</text>')
        lines.append(indent +  '</message>')
        self._report.write("\n".join(lines) + "\n")


    def passed(self, text: str, details: str):
        """
        Write tags for a successful verification
        """
        indent = 6 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent +  '<verification>')
        lines.append(
            indent + f' <scriptedVerificationResult type="PASS" time="{timestamp}">')
        lines.append(indent + f'  <text>{escape(text)}</text>')
        lines.append(indent + f'  <detail>{escape(details)}</detail>')
        lines.append(indent +  ' </scriptedVerificationResult>')
        lines.append(indent +  '</verification>')
        self._report.write("\n".join(lines) + "\n")


    def failed(self, text: str, details: str, trace = None):
        """
        Write tags for a failed verification
        """
        max_nb_frames = 5
        if trace is None:
            summary = tb.StackSummary.extract(tb.walk_stack(None))
            # Skip current frame
            start_index = 1
        else:
            summary = tb.StackSummary.extract(tb.walk_tb(trace))
            # Skip Behave frames
            start_index = 2 if len(summary) > 2 else 0
        indent = 6 * " "
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        lines = []
        lines.append(indent +  '<verification>')
        # Stacktrace
        lines.append(indent +  ' <stacktrace>')
        for frame in summary[start_index:start_index + max_nb_frames]:
            lines.append(indent +  '  <frame>')
            lines.append(indent +  '   <location>')
            lines.append(indent + f'    {frame.filename}:{frame.lineno}:{frame.name}')
            lines.append(indent +  '   </location>')
            lines.append(indent +  '   <line>')
            lines.append(indent + f'    {frame.line}')
            lines.append(indent +  '   </line>')
            lines.append(indent +  '  </frame>')
        lines.append(indent +  ' </stacktrace>')
        # Results
        lines.append(
            indent + f' <scriptedVerificationResult type="FAIL" time="{timestamp}">')
        lines.append(indent + f'  <text>{escape(text)}</text>')
        lines.append(indent + f'  <detail>{escape(details)}</detail>')
        lines.append(indent +  ' </scriptedVerificationResult>')
        lines.append(indent +  '</verification>')
        self._report.write("\n".join(lines) + "\n")



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Classes and functions to manage application lifecycle
"""

from importlib import resources as res
from pathlib import Path
from typing import Optional
import json
import os
import platform
import subprocess
import sys
import tempfile
import time

import qat.internal.application_context as app_ctxt
from qat.internal import debug_operations
from qat.test_settings import Settings

import qat.test_settings


def is_linux() -> bool:
    """
    Return whether the current platform is Linux or not
    """
    system = platform.system()
    return system == "Linux"


def is_windows() -> bool:
    """
    Return whether the current platform is Windows or not
    """
    system = platform.system()
    return system == "Windows"


def get_temp_folder() -> str:
    """
    Return the path to the temporary folder
    """

    if 'TEMP' not in os.environ:
        os.environ['TEMP'] = tempfile.gettempdir()
    return os.environ['TEMP']


def is_debugging() -> bool:
    """
    Return whether the debugger is currently active or not
    """
    is_debugger_active = hasattr(sys, 'gettrace') and sys.gettrace() is not None
    if not is_debugger_active:
        return False
    if 'COVERAGE' in os.environ:
        return len(os.environ['COVERAGE']) == 0
    return is_debugger_active


def get_injector(app_path: Path) -> Path:
    """
    Return the path to the Injector application.
    """
    if is_windows():
        injector_name = 'injector.exe'
    elif is_linux():
        injector_name = 'libinjector.so'
    else:
        raise NotImplementedError('Current platform is not supported')

    # Find injector in installed package
    try:
        injector_path = res.files('qat') / 'bin'
        injector_exe = injector_path / injector_name
        if os.path.exists(injector_exe):
            return injector_exe
    except: # pylint: disable = bare-except
        print("Warning: Qat requires Python >= 3.9")

    # Find injector in dev environment
    for root_path in [
        Path(os.getcwd()) / 'client' / 'qat' / 'bin',
        Path(os.getcwd()) / 'build']:
        matches = list(root_path.glob(f"**/{injector_name}"))
        if len(matches) > 0:
            return matches[0]

    # Last try: get locally installed binaries
    injector_exe = app_path / injector_name
    if not os.path.exists(injector_exe):
        raise FileNotFoundError("Could not find Injector binaries")

    return injector_exe


def inject_dll(context: app_ctxt.ApplicationContext):
    """
    Call Injector application to load the qat server DLL into the target application
    """
    injector_exe = get_injector(context.get_path())
    injector_path = injector_exe.parent
    try:
        subprocess.run(
            [injector_exe, str(context.pid), str(
                injector_path / "injector.dll")],
            cwd=injector_path,
            shell=False,
            check=True,
            timeout=Settings.wait_for_app_start_timeout)
    except subprocess.TimeoutExpired as error:
        # Injector may have run too early
        context.kill()
        raise TimeoutError("Application could not start (timeout)") from error
    except:
        context.kill()
        raise


def connect_to(context: app_ctxt.ApplicationContext):
    """
    Establish a TCP communication with the target application
    """
    # Retrieve server port from config file
    context.config_file = Path(get_temp_folder()) / f"qat-{context.pid}.txt"
    port = -1
    delay = 0.2
    nb_tries = 10 * int(1/delay)  # 10 seconds
    last_error = None
    for _ in range(nb_tries):
        try:
            last_error = None
            with open(context.config_file, 'rt', encoding="utf-8") as file:
                content = file.readline()
                content = content.strip()
                port = int(content)
                break
        except (OSError, ValueError) as error:
            time.sleep(delay)
            # Handle cases when the application exits prematurely
            if context.is_finished():
                print('Abort: app terminated')
                break
            last_error = error
    if port < 0:
        if last_error is not None:
            print(f"Error reading pid file {last_error}")
        context.kill()
        raise ProcessLookupError(
            "Could not retrieve server port number. "
            "Cannot establish communication with application.")

    # Establish TCP communication
    context.init_comm(port)
    # Add version info to context
    context.init_version_info()


def attach_to(name_or_pid) -> app_ctxt.ApplicationContext:
    """
    Attach to the given application by name or process ID.
    If a name is given, it must correspond to a registered application.
    """
    qat.test_settings.load_test_settings()
    timeout = Settings.wait_for_app_start_timeout
    start_time = round(1000 * time.time())
    if isinstance(name_or_pid, str) and len(name_or_pid) > 0:
        if name_or_pid.endswith('.exe'):
            name_or_pid = name_or_pid[0:-4]
        if is_linux():
            args = ['pidof', '-s', f'{name_or_pid}']
        elif is_windows():
            command = f'Get-Process -Name {name_or_pid} | Select -ExpandProperty Id'
            args = ['powershell', '-command', command]
        found = False
        while not found and (round(1000 * time.time()) - start_time) < timeout:
            try:
                result = subprocess.check_output(
                    args, text=True, shell=False)
                pid = int(result.strip())
                found = True
            except (ValueError, subprocess.CalledProcessError):
                time.sleep(0.2)

        if not found:
            raise ProcessLookupError(f"Could not find process '{name_or_pid}'")

    # 'bool' is a subclass of 'int' in Python
    elif isinstance(name_or_pid, int) and not isinstance(name_or_pid, bool):
        pid = name_or_pid
    else:
        raise ValueError(
            "Invalid argument given to attach_to_application(): must be app name or process ID")

    if is_linux():
        args = ['readlink', '-f', f'/proc/{pid}/exe']
    elif is_windows():
        command = f'Get-Process -Id {pid} -FileVersionInfo | Select -ExpandProperty FileName'
        args = ['powershell', '-command', command]

    found = False
    while not found and (round(1000 * time.time()) - start_time) < timeout:
        try:
            result = subprocess.check_output(args, text=True, shell=False)
        except subprocess.CalledProcessError:
            result = ''
        paths = result.split('\n')
        found = len(paths) > 0 and len(paths[0]) > 0
        if not found:
            time.sleep(0.2)

    if not found:
        raise ProcessLookupError(f"Could not find process '{name_or_pid}'")
    app_path = Path(paths[0]).resolve()
    context = app_ctxt.ApplicationContext(name_or_pid, app_path)
    context.pid = pid

    if is_windows():
        inject_dll(context)
    connect_to(context)

    return context


def get_config_file(shared: Optional[bool] = None) -> Path:
    """
    Return the current config file
    """
    local_file = Path(os.getcwd()) / 'applications.json'
    shared_file = Path.home() / '.qat' / 'applications.json'

    # Select local file if it exists or if shared file does not
    if shared is None:
        if local_file.is_file():
            config_file = local_file
        elif shared_file.is_file():
            config_file = shared_file
        else:
            config_file = local_file
    elif shared:
        config_file = shared_file
    else:
        config_file = local_file

    return config_file


def list_applications() -> dict:
    """
    List all registered applications (shared and local)
    """
    applications = {}
    # Load shared applications first
    try:
        with open(get_config_file(True), 'r', encoding='utf-8') as file:
            applications = json.load(file)
    except: # pylint: disable=bare-except
        pass

    # Load local applications, overwriting shared ones in case of conflict
    try:
        with open(get_config_file(False), 'r', encoding='utf-8') as file:
            applications.update(json.load(file))
    except: # pylint: disable=bare-except
        pass
    return applications


def cleanup_temp_files():
    """
    Delete old temporary files
    """
    temp_folder = Path(get_temp_folder())
    for file in temp_folder.glob('qat-*.txt'):
        pid = file.stem[4:]
        if is_linux():
            try:
                with subprocess.Popen(
                    ['ps', '-e'], stdout=subprocess.PIPE, text=True, shell=False) as ps_proc:
                    result = subprocess.check_output(
                        ['grep', '-w', f'{pid}'], stdin=ps_proc.stdout, text=True, shell=False)

            except subprocess.CalledProcessError:
                continue
        elif is_windows():
            command = f'Get-Process -Id {pid} | Select -ExpandProperty ProcessName'
            try:
                result = subprocess.run(
                    ['powershell', '-command', command],
                    text=True,
                    shell=False,
                    check=False,
                    stdout=subprocess.PIPE).stdout
            except subprocess.CalledProcessError:
                continue
        result = result.strip()
        if len(result) == 0:
            try:
                os.remove(file)
            except OSError:
                print(f'Could not delete temporary file {file}')


def start_application(app_name: str, args: str, detached=False) -> app_ctxt.ApplicationContext:
    """
    Start the given application, inject the server library (except if detached is True)
    and return the corresponding application context
    """

    if not (is_linux() or is_windows()):
        raise NotImplementedError(f"{platform.system()} is unsupported.")

    qat.test_settings.load_test_settings()
    try:
        apps = list_applications()
        app_path = apps[app_name]['path']
        if not app_path or not os.path.exists(app_path):
            raise FileNotFoundError("Invalid application path")
        app_path = Path(app_path).resolve()
        if not app_path.is_file():
            app_path /= app_name + '.exe'
        if not app_path.exists():
            raise ValueError(f"Application '{app_name}' does not exist")
        if args is None:
            args = apps[app_name]['args']

    except Exception as error:
        raise ValueError(
            f"Application '{app_name}' is not defined in "
            "configuration file 'applications.json'") from error

    context = app_ctxt.ApplicationContext(app_name, app_path)

    cleanup_temp_files()

    # Prepare environment
    local_env = dict(os.environ)
    if 'TEMP' not in local_env:
        local_env['TEMP'] = get_temp_folder()

    if is_linux():
        local_env['LD_PRELOAD'] = str(get_injector(app_path.parent))

    # Launch application
    try:
        context.launch(args, local_env, is_linux())

        if detached:
            return context

        if is_windows():
            inject_dll(context)

        connect_to(context)

        # Lock GUI
        lock_ui = Settings.lock_ui.lower()
        if lock_ui == "always" or (lock_ui == "auto" and not is_debugging()):
            debug_operations.lock_application(context)

    except Exception as error:
        print("Failed to connect to application. Stopping process...")
        context.kill()
        print(f"Application {app_name} was killed.")
        raise error

    print(f"Application {app_name} successfully started and ready for testing")
    return context



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Functions related to gestures
"""

from decimal import Decimal
import decimal
import math
import time

from qat.internal.application_context import ApplicationContext
from qat.internal import find_object
from qat.internal import touch_operations


def round_float_to_int(value: float) -> int:
    """
    Return the given value rounded to the nearest integer (using ROUND_HALF_UP strategy)
    """
    decimal.getcontext().rounding = decimal.ROUND_HALF_UP
    return int(round(Decimal(str(value)), 0))


def flick(
        app_context: ApplicationContext,
        definition: dict,
        dx=0,
        dy=0):
    """
    Move the given Flickable by the given horizontal and vertical distances in pixels.
    """
    definition = find_object.object_to_definition(definition)
    find_object.wait_for_object(app_context, definition)

    args = {
        'dx': dx,
        'dy': dy
    }

    command = {}
    command['command'] = 'gesture'
    command['object'] = definition
    command['attribute'] = 'flick'
    command['args'] = args

    app_context.send_command(command)


def pinch(
        app_context: ApplicationContext,
        definition: dict,
        rotation = 0.,
        translation = None,
        scale = 1.0):
    """
    Generate a pinch event (zoom, rotation, pan).
    """
    if scale <= 0:
        raise ValueError("Scale argument must be a strict positive number")

    if translation is None:
        translation = [0, 0]

    definition = find_object.object_to_definition(definition)
    widget = find_object.wait_for_object(app_context, definition)
    w = widget.width
    h = widget.height

    initial_rotation = 0
    final_rotation = rotation

    # Compute initial and final positions
    initial_position = [(w - translation[0]) / 2, (h - translation[1]) / 2]

    if initial_position[0] <= 0 or initial_position[1] <= 0:
        raise ValueError("Translation is out of widget's boundaries")

    final_position = [sum(element) for element in zip(initial_position, translation)]

    # Scale the gesture depending on the widget's size
    available_distance = min(initial_position[0], initial_position[1], w - initial_position[0], h - initial_position[1])
    if scale < 1:
        initial_scale = 0.9 * available_distance
        final_scale = scale * initial_scale
    else:
        final_scale = 0.9 * available_distance
        initial_scale = final_scale / scale

    # Decompose the gesture into smaller steps
    num_steps = 10
    rotation_step = (final_rotation - initial_rotation) / num_steps
    horizontal_step = (final_position[0] - initial_position[0]) / num_steps
    vertical_step = (final_position[1] - initial_position[1]) / num_steps
    scale_step = (final_scale - initial_scale) / num_steps

    x1 = round_float_to_int(initial_position[0] - initial_scale)
    y1 = round_float_to_int(initial_position[1])
    x2 = round_float_to_int(initial_position[0] + initial_scale)
    y2 = round_float_to_int(initial_position[1])

    try:
        # Qt needs a minimal drag distance to initiate a gesture, so
        # send the first event a bit off on the X axis. That way, the first update
        # will start the gesture at the desired position. This has no effect on the
        # gesture itself.
        # see QStyleHints::startDragDistance
        # todo: use the actual value since it is configurable and OS-dependent
        minimum_drag_distance = 10
        touch_operations.tap(
            app_context,
            definition,
            "press",
            x = [x1 + minimum_drag_distance, x2 - minimum_drag_distance],
            y = [y1, y2])

        for i in list(range(num_steps + 1)):
            # Add a small delay to create a more realistic gesture
            time.sleep(0.02)
            center_x = initial_position[0] + i * horizontal_step
            center_y = initial_position[1] + i * vertical_step
            angle = initial_rotation + i * rotation_step
            angle = math.radians(-angle)
            length = initial_scale + i * scale_step
            x1 = round_float_to_int(center_x - length * math.cos(angle))
            y1 = round_float_to_int(center_y + length * math.sin(angle))
            x2 = round_float_to_int(center_x + length * math.cos(angle))
            y2 = round_float_to_int(center_y - length * math.sin(angle))

            touch_operations.tap(
                app_context,
                definition,
                "move",
                x = [x1, x2],
                y = [y1, y2])
    finally:
        touch_operations.tap(
            app_context,
            definition,
            "release",
            x = [x1, x2],
            y = [y1, y2])


def native_pinch(
        app_context: ApplicationContext,
        definition: dict,
        angle: float = None,
        scale: float = None,
        check: bool = False):
    """
    Generate a native pinch event (zoom and/or rotation).
    """
    definition = find_object.object_to_definition(definition)

    # Handle optional arguments
    args = {}
    if angle is not None:
        args['angle'] = angle
    if scale is not None:
        args['scale'] = scale

    command = {}
    command['command'] = 'gesture'
    command['object'] = definition
    command['attribute'] = 'pinch'
    command['args'] = args

    result = app_context.send_command(command)

    if check and 'warning' in result:
        raise RuntimeWarning(result['warning'])



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Provides the QtMethod class
"""

# Avoid cyclic imports
# pylint: disable = consider-using-from-import
# pylint: disable = cyclic-import

from dataclasses import dataclass
from pathlib import Path
import json

import qat.internal.qt_object as qt_object
import qat.internal.qt_custom_object as qt_custom_object

@dataclass
class QtMethod():
    """
    Bind a remote Qt method to a local object
    """

    def __init__(self, qobject, name: str) -> None:
        """
        Constructor the current value of a Qt variable
        """
        if not isinstance(qobject, qt_object.QtObject):
            raise ValueError(
                f"Cannot create Method {name}: no valid QtObject given")
        self._object = qobject
        self._name = name

    def __call__(self, *args):
        """
        Call this method with the given arguments.
        """
        args = list(args)
        for index, item in enumerate(args):
            if isinstance(item, qt_custom_object.QtCustomObject):
                args[index] = item.get_attribute_dict()
            elif isinstance(item, qt_object.QtObject):
                args[index] = item.get_definition()
            elif isinstance(item, Path):
                args[index] = str(item)
            else:
                args[index] = item

        command = {}
        command['command'] = 'call'
        command['object'] = self._object.__dict__['_definition']
        command['attribute'] = self._name
        command['args'] = args

        result = self._object._app_context.send_command(command)

        if 'value' in result:
            value = result['value']
            if 'returnValue' in value:
                value = value['returnValue']
                if isinstance(value, dict):
                    return qt_custom_object.QtCustomObject(value)
                return value
            if 'returnObject' in value:
                value = json.loads(value)
                return qt_object.QtObject(
                    self._object.__dict__['_app_context'],
                    value['returnObject'])

        raise RuntimeError(f"Calling '{self._name}' did not return any value")



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Provides the SpyWindow class
"""

import tkinter as tk
from tkinter import ttk
from qat.gui import gui_utils
import qat


class SpyWindow():
    """
    Main window of Qat Spy when connected to an application
    """
    # pylint: disable=attribute-defined-outside-init
    def __init__(self, root: tk.Tk, start_window) -> None:
        self._root = root
        self._start_window = start_window
        self.spy_window = None
        self.is_picking = False
        self._selected_object = None
        self.main_windows_list = []
        self.current_root_objects = []


    def generate_spy_window(self, image_dir):
        """
        Create components of the main window
        """
        self.spy_window = tk.Toplevel(self._root)
        self.spy_window.title('Qat Spy')
        self.spy_window.geometry('1000x600')
        self.spy_window.protocol("WM_DELETE_WINDOW", self.back_to_start)
        self.spy_window.withdraw()

        # Specify Grid
        tk.Grid.columnconfigure(self.spy_window, 0, weight=1)
        tk.Grid.columnconfigure(self.spy_window, 1, weight=1)
        tk.Grid.columnconfigure(self.spy_window, 2, weight=1)
        tk.Grid.columnconfigure(self.spy_window, 3, weight=1)
        tk.Grid.columnconfigure(self.spy_window, 4, weight=1)
        tk.Grid.rowconfigure(self.spy_window, 1, weight=1)

        # define columns
        columns = 'type'

        self.tree = ttk.Treeview(
            self.spy_window, columns=columns, selectmode='browse')

        # default column is accessed using #0
        self.tree.heading('#0', text="Object")
        self.tree.heading('type', text='Type')

        self.tree.bind('<<TreeviewOpen>>', self.item_opened)
        self.tree.bind('<<TreeviewSelect>>', self.item_selected)

        self.tree.grid(row=1, column=0, sticky='nsew')

        # add a scrollbar
        self.scrollbar = ttk.Scrollbar(
            self.spy_window, orient=tk.VERTICAL, command=self.tree.yview)
        self.tree.configure(yscroll=self.scrollbar.set)
        self.scrollbar.grid(row=1, column=1, sticky='ns')

        self.properties_tree = ttk.Treeview(
            self.spy_window, columns=("value"), selectmode='browse')
        # define headings
        self.properties_tree.heading('#0', text='Property')
        self.properties_tree.heading('value', text='Value')

        self.properties_tree.grid(row=1, column=2, sticky='nsew')

        # add a scrollbar
        vert_scrollbar = ttk.Scrollbar(
            self.spy_window, orient=tk.VERTICAL, command=self.properties_tree.yview)
        self.properties_tree.configure(yscroll=vert_scrollbar.set)
        vert_scrollbar.grid(row=1, column=3, sticky='ns')

        # Add toolbar
        self.toolbar = tk.Frame(self.spy_window, bd=1, relief=tk.RAISED)
        self.toolbar.grid(row=0, column=0, columnspan=3, sticky='we')

        # add picker button
        self.picker_icon = tk.PhotoImage(file=image_dir / "spy_icon.png")
        self.picker_button = tk.Button(
            self.toolbar, text="Picker", image=self.picker_icon, command=self.pick_item)
        self.picker_button.grid(row=0, column=0, sticky='w')

        # add up button
        self.up_icon = tk.PhotoImage(
            file=image_dir / "up_icon.png", master=self._root)
        self.up_button = tk.Button(
            self.toolbar, text="Up one level", image=self.up_icon, command=self.go_up)
        self.up_button.grid(row=0, column=1, sticky='w')

        # add refresh button
        self.refresh_icon = tk.PhotoImage(file=image_dir / "refresh_icon.png")
        self.refresh_button = tk.Button(
            self.toolbar, text="Refresh", image=self.refresh_icon, command=self.refresh)
        self.refresh_button.grid(row=0, column=2, sticky='w')

        self.pick_label = tk.Label(
            self.toolbar,
            text='While picking, hold CTRL to interact with app or SHIFT to pick background object',
            bd=0)
        self.pick_label.grid(row=0, column=3, sticky='w')

        # Create a tooltip for each buttons of the toolbar
        gui_utils.create_button_tooltip(self.toolbar, 1.0)


    def show(self):
        """
        Display the main window
        """
        self.spy_window.deiconify()
        self.is_picking = False


    def back_to_start(self):
        """
        Close current application and display the Application Manager window
        """
        try:
            qat.close_application()
        except: # pylint: disable=bare-except
            print("Could not close application")

        self.spy_window.withdraw()
        self._start_window.deiconify()


    def go_up(self):
        """
        Select the parent object in the tree
        """
        # If no item is selected choose the first item in tree
        if len(self.tree.selection()) > 0:
            selected_id = self.tree.selection()[0]
        else:
            selected_id = self.tree.get_children('')[0]

        parent = self.tree.parent(selected_id)
        if parent:
            self.tree.focus(parent)
            self.tree.selection_set(parent)
        else:
            selected_object = self.ids[selected_id]
            try:
                parent_object = selected_object.parent
                if parent_object.get_definition() is not None:
                    self.populate_tree(parent_object)
                else:
                    self.populate_tree([])
            except Exception: # pylint: disable=broad-exception-caught
                pass


    def populate_tree(self, root_object_list):
        """
        Populate the object tree from the given root objects
        """
        if not isinstance(root_object_list, list):
            root_object_list = [root_object_list]

        self.current_root_objects = root_object_list
        if len(root_object_list) == 0:
            root_object_list = qat.list_top_windows()

        self.tree.delete(*self.tree.get_children())

        self.ids = {}

        for root_object in root_object_list:
            root_id = self.tree.insert('', tk.END, text=str(
                root_object), values=(root_object.className,), open=False)
            self.ids[root_id] = root_object

            if len(root_object.children) > 0:
                self.tree.insert(root_id, index=0, text="loading",
                                 values=("loading",), open=False)

        self.refresh_properties()


    def refresh_properties(self):
        """
        Refresh all values in the property table
        """
        if len(self.tree.selection()) == 0:
            self.properties_tree.delete(*self.properties_tree.get_children())
            return
        selected_object = self.ids[self.tree.selection()[0]]
        self.populate_properties(selected_object)


    def refresh(self):
        """
        Refresh all elements and preserve selection if possible
        """
        self.populate_tree(self.current_root_objects)


    def populate_properties(self, selected_object):
        """
        Populate the property table for the given object
        """
        # pylint: disable=bare-except
        self.properties_tree.delete(*self.properties_tree.get_children())
        properties = self.sort_properties(selected_object.list_properties())
        for prop in properties:
            try:
                value = prop[1]
                tree_id = self.properties_tree.insert(
                    '', tk.END, text=prop[0], values=(str(value),), open=False)
                if isinstance(value, qat.QtObject):
                    self.add_object(tree_id, value)
                elif isinstance(value, qat.QtCustomObject):
                    self.add_custom_object(tree_id, value)

            except:
                pass


    def add_object(self, tree_id, qt_object: qat.QtObject):
        """
        Add the given object to the property table
        """
        properties = self.sort_properties(qt_object.list_properties())
        for prop in properties:
            try:
                property_value = prop[1]
                child_id = self.properties_tree.insert(
                    tree_id,
                    tk.END,
                    text=prop[0],
                    values=(str(property_value),), open=False)
                if isinstance(property_value, qat.QtCustomObject):
                    self.add_custom_object(child_id, property_value)

            except Exception: # pylint: disable=broad-exception-caught
                pass


    def add_custom_object(self, tree_id, custom_object: qat.QtCustomObject):
        """
        Add the given custom object to the property table
        """
        for prop in custom_object:
            try:
                property_value = custom_object[prop]
                self.properties_tree.insert(
                    tree_id,
                    tk.END,
                    text=prop,
                    values=(str(property_value),),
                    open=False)
            except: # pylint: disable=bare-except
                pass


    def sort_properties(self, properties: list) -> list:
        """
        Sort given properties to move most important ones to the beginning of the list
        """
        properties = dict(properties)
        properties = dict(sorted(properties.items()))
        # Parent and children are in the tree
        if 'children' in properties:
            del properties['children']
        if 'parent' in properties:
            del properties['parent']
        top_properties_names = ['objectName', 'type', 'id']
        top_properties = {}
        for top_prop in top_properties_names:
            if top_prop in properties:
                top_properties[top_prop] = properties[top_prop]
                del properties[top_prop]

        return list(top_properties.items()) + list(properties.items())


    def item_opened(self, _):
        """
        Callback when an item is opened in the tree.
        Populate children of the expanded object.
        """
        selected_id = self.tree.selection()[0]
        selected_children_ids = self.tree.get_children(selected_id)
        if len(selected_children_ids) > 1:
            # Tree has already been expanded
            return
        for temp_id in selected_children_ids:
            self.tree.delete(temp_id)
        children = self.ids[selected_id].children
        for child in reversed(children):
            if child.type == 'RootItem':
                continue
            tree_id = self.tree.insert(selected_id, index=0, text=str(
                child), values=(child.className,), open=False)
            self.ids[tree_id] = child
            if len(child.children) > 0:
                self.tree.insert(tree_id, index=0, text="loading",
                                 values=("loading",), open=False)


    def item_selected(self, _):
        """
        Callback when an item is selected in the tree.
        Populate the property table.
        """
        if len(self.tree.selection()) == 0:
            self.properties_tree.delete(*self.properties_tree.get_children())
            self._selected_object = None
            return
        selected_object = self.ids[self.tree.selection()[0]]
        if selected_object == self._selected_object:
            return
        self._selected_object = selected_object
        self.populate_properties(selected_object)


    def pick_item(self):
        """
        Toggle picking mode
        """
        self.is_picking = not self.is_picking
        self.picker_button.config(
            relief="sunken" if self.is_picking else "raised")
        self.up_button["state"] = tk.DISABLED if self.is_picking else tk.NORMAL
        if self.is_picking:
            self.wait_for_picker()
        else:
            try:
                for conn in self._picker_connections:
                    qat.disconnect(conn)
                self._picker_connections = []
                qat.deactivate_picker()
            except: # pylint: disable=bare-except
                print('error while deactivating picker')


    def wait_for_picker(self):
        """
        Connect to the object picker to retrieve the picked object
        """
        def set_picked_object(picked_object):
            self.populate_tree(picked_object)
            first_id = self.tree.get_children('')[0]
            self.tree.focus(first_id)
            self.tree.selection_set(first_id)
            self.picker_button.invoke()

        if self.is_picking:
            try:
                qat.activate_picker()
                picker_def = {
                    'objectName': 'QatObjectPicker'
                }

                pickers = qat.find_all_objects(picker_def)
                self._picker_connections = []
                for picker in pickers:
                    self._picker_connections.append(qat.connect(
                        picker, 'pickedObject', set_picked_object))
            except: # pylint: disable=bare-except
                print('error while activating picker')
                self.picker_button.invoke()



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Provides the QtObject class
"""

from copy import deepcopy
import json
import os

import qat.internal.application_context as app_ctxt
import qat

# Avoid cyclic imports
# pylint: disable = consider-using-from-import
import qat.internal.qt_method as qt_method
from qat.internal.qt_custom_object import QtCustomObject


class QtObject():
    """
    Local representation of a remote Qt object
    """

    def __init__(self, context: app_ctxt.ApplicationContext, definition: dict) -> None:
        """
        Store the application context and a copy of the object's definition
        """
        self.__dict__['_definition'] = deepcopy(definition)
        self.__dict__['_app_context'] = context


    def __del__(self):
        if '_definition' in self.__dict__ and self.__dict__['_definition'] is not None:
            if '_temp_image_file' in self.__dict__['_definition']:
                tmp_file = self.__dict__['_definition']['_temp_image_file']
                if os.path.exists(tmp_file):
                    try:
                        os.remove(tmp_file)
                    except OSError as error:
                        print("Could not delete temp file: " + str(error))


    def get_definition(self) -> dict:
        """
        Return the definition of this object as a dictionary of properties
        """
        return self.__dict__['_definition']


    def is_null(self) -> bool:
        """
        Return whether this object is a NULL pointer or not
        """
        if self.__dict__['_definition'] is None:
            return True
        if 'cache_uid' not in self.__dict__['_definition']:
            return False
        return self.__dict__['_definition']['cache_uid'] == '0'


    def property(self, name: str):
        """
        Get the remote attribute of this object.
        """
        return getattr(self, name)


    def _get_app_context(self) -> app_ctxt.ApplicationContext:
        context = self.__dict__['_app_context']
        if context is None or not context.is_running():
            # Context was lost. Use current context
            context = qat.current_application_context()
            if context is None or not context.is_running():
                return None
            self.__dict__['_app_context'] = qat.current_application_context()
            definition = self.__dict__['_definition']
            if definition is not None and 'cache_uid' in definition:
                definition['cache_uid'] = self.cache_uid
        return self.__dict__['_app_context']


    def __str__(self):
        """
        Custom string representation of a QtObject: objectName, id or <unnamed>
        """
        definition = self.__dict__['_definition']
        if 'objectName' in definition:
            return definition['objectName']
        if 'id' in definition:
            return definition['id']
        text = self.objectName
        if text == "":
            try:
                text = self.id
            except AttributeError:
                pass
            if text == "":
                text = '<unnamed>'
        return str(text)


    def list_properties(self) -> list:
        """
        Return the list of all available properties for this object.
        """
        command = {}
        command['command'] = 'list'
        command['object'] = self.__dict__['_definition']
        command['attribute'] = "properties"

        try:
            result = self._get_app_context().send_command(command)
        except LookupError as error:
            raise AttributeError(str(error)) from error


        property_list = []
        object_tag = "<object>:"
        if 'properties' in result:
            for name in result['properties']:
                value = result['properties'][name]
                if isinstance(value, str) and value.startswith(object_tag):
                    object_def = json.loads(value[len(object_tag):])
                    value = QtObject(self.__dict__['_app_context'], object_def)
                    if value.is_null():
                        value = "<null>"
                elif isinstance(value, dict):
                    value = QtCustomObject(value)
                property_list.append((name, value))
            return property_list

        raise RuntimeError("Unknown error while listing properties")


    def __getattr__(self, name: str):
        """
        Get the remote attribute of this object.
        """
        command = {}
        command['command'] = 'get'
        command['object'] = self.__dict__['_definition']
        command['attribute'] = name

        try:
            result = self._get_app_context().send_command(command)
        except RuntimeError as error:
            raise AttributeError(str(error)) from error

        # Objects such as ImageWrapper can be re-generated dynamically so
        # update their cache UID automatically.
        if 'cache_uid' in self.__dict__['_definition'] and 'cache_uid' in result:
            self.__dict__['_definition']['cache_uid'] = result['cache_uid']

        if 'value' in result:
            # Parse attribute value
            json_value = result['value']
            if isinstance(json_value, dict):
                return QtCustomObject(json_value)
            return json_value
        if 'object' in result:
            # Parse object definition
            return QtObject(self.__dict__['_app_context'], result['object'])
        if 'children' in result:
            # Build children
            children = []
            for child in result['children']:
                children.append(QtObject(self.__dict__['_app_context'], child))
            return children
        if 'found' in result and result['found']:
            # Attribute is a method
            return qt_method.QtMethod(self, name)

        raise AttributeError(f"Unknown error while getting attribute '{name}'")


    def __setattr__(self, name: str, value):
        """
        Set the remote attribute of this object.
        """

        if isinstance(value, QtCustomObject):
            value = value.get_attribute_dict()

        command = {}
        command['command'] = 'set'
        command['object'] = self.__dict__['_definition']
        command['attribute'] = name
        command['args'] = value

        try:
            self._get_app_context().send_command(command)
        except RuntimeError as error:
            raise AttributeError(str(error)) from error


    def __eq__(self, other):
        if other is None:
            return self.is_null()
        if isinstance(other, dict):
            other = qat.wait_for_object_exists(other)

        return self.cache_uid == other.cache_uid



# -*- coding: utf-8 -*-

"""
Definitions of all objects used by the tests
"""

# Define each object in a property dictionary
main_window = {
    'type': 'ApplicationWindow'
}

# Map a unique name to each object definition.
# These names will be used in feature files.
bdd_mapping = {
    'Main Window': main_window
}



# -*- coding: utf-8 -*-

"""
When steps implementation
"""

from behave import When

from qat import report
import qat

@When("This application is started")
def step(context):
    app_name = context.userData['app_name']
    # Save application context to context so other steps can use it later
    context.userData['app_context'] = qat.start_application(app_name)
    report.log(f"'{app_name}' was successful started")




# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Functions related to mouse
"""

from dataclasses import dataclass

from qat.internal.application_context import ApplicationContext
from qat.test_settings import Settings
from qat.internal import find_object

# Mouse events require many arguments
# pylint: disable = too-many-arguments

@dataclass
class Button():
    """
    Mouse button constants
    """
    NONE = 'none'
    LEFT = 'left'
    RIGHT = 'right'
    MIDDLE = 'middle'


@dataclass
class Modifier():
    """
    Key modifier constants
    """
    NONE = 'none'
    ALT = 'alt'
    CTL = 'ctrl'
    SHIFT = 'shift'


def click(
        app_context: ApplicationContext,
        definition: dict,
        mode: str,
        x=None,
        y=None,
        modifier=Modifier.NONE,
        button=Button.LEFT,
        check=False):
    """
    Click, double-click or move the mouse on the given widget using the given parameters.
    """
    definition = find_object.object_to_definition(definition)
    find_object.wait_for_object(app_context, definition)

    args = {
        'button': button,
        'modifier': modifier
    }
    if x is not None and y is not None:
        args['x'] = x
        args['y'] = y
    command = {}
    command['command'] = 'mouse'
    command['object'] = definition
    command['attribute'] = mode
    command['args'] = args

    result = app_context.send_command(command)

    if check and 'warning' in result:
        raise RuntimeWarning(result['warning'])


def drag(
        app_context: ApplicationContext,
        definition: dict,
        mode: str,
        x=None,
        y=None,
        dx=0,
        dy=0,
        modifier=Modifier.NONE,
        button=Button.LEFT,
        check=False):
    """
    Drag the mouse using the given parameters.
    """
    definition = find_object.object_to_definition(definition)
    find_object.wait_for_object(app_context, definition)

    args = {
        'button': button,
        'modifier': modifier
    }
    if x is not None and y is not None:
        args['x'] = x
        args['y'] = y
    args['dx'] = dx
    args['dy'] = dy
    command = {}
    command['command'] = 'mouse'
    command['object'] = definition
    command['attribute'] = mode
    command['args'] = args

    result = app_context.send_command(command, Settings.long_operation_timeout)

    if check and 'warning' in result:
        raise RuntimeWarning(result['warning'])



# -*- coding: utf-8 -*-

"""
Given steps implementation
"""

from behave import Given
from qat import report
import qat

@Given("An application is registered")
def step(context):
    registered_apps = qat.list_applications()
    if len(registered_apps) == 0:
        report.failed(
            'No application found',
            'Use qat-gui command or qat.register_application() to register an application'
        )
    app_name = list(registered_apps)[0]
    report.log(f"'{app_name}' is available")
    # Add application name to context so other steps can access it
    context.userData['app_name'] = app_name



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Hooks intended to be called from environment.py when using Behave
"""

from pathlib import Path
import os
from behave.model import Scenario, Status

from qat.report import PropagatedException
from qat import test_settings
import qat


def before_all(context, suite_name):
    """
    Called before a test starts.
    Setup XML report file and initialize test settings from JSON file.
    """
    if 'report_file' in context.config.userdata:
        report_filename = Path(context.config.userdata['report_file'])
    else:
        report_filename = Path(os.getcwd()) / 'report' / 'results.xml'

    qat.Globals.current_report = qat.XmlReport(suite_name, report_filename)

    # Apply settings
    test_settings.load_test_settings()
    Scenario.continue_after_failed_step = test_settings.Settings.continue_after_fail


def before_feature(context, feature):
    """
    Called before a feature starts.
    Add an entry to the current report.
    """
    context.userData = {}
    test_name = Path(feature.filename).resolve().parent.stem
    qat.Globals.current_report.start_test_case(test_name)
    qat.Globals.current_report.start_feature(
        feature.name, '\n'.join(feature.description))


def before_scenario(_, scenario):
    """
    Called before a scenario starts.
    Add an entry to the current report.
    """
    qat.Globals.current_report.start_scenario(
        scenario.name, '\n'.join(scenario.description))
    # pylint: disable = protected-access
    if hasattr(scenario, '_row') and scenario._row is not None:
        qat.Globals.current_report.start_example(scenario._row.headings, scenario._row.cells)


def before_step(context, step):
    """
    Called before a step starts.
    Add an entry to the current report.
    """
    qat.Globals.current_report.start_step(step.name)
    if hasattr(step, 'text') and step.text is not None:
        context.multiLineText = step.text.splitlines()


def after_step(_, step):
    """
    Called after a step ends.
    Add an entry to the current report.
    """
    if step.status == Status.failed:
        if not isinstance(step.exception, PropagatedException):
            qat.Globals.current_report.failed(
                "Script Error", str(step.exception), step.exc_traceback)
    qat.Globals.current_report.end_step()


def after_scenario(*_):
    """
    Called after a scenario ends.
    Add an entry to the current report and close the current application.
    """
    qat.Globals.current_report.end_scenario()
    try:
        qat.close_application()
    except Exception as error:  # pylint: disable = broad-exception-caught
        print("Could not close application: " + str(error))


def after_feature(*_):
    """
    Called after a feature ends.
    Add an entry to the current report and closes the report.
    """
    qat.Globals.current_report.end_feature()
    qat.Globals.current_report.end_test_case()


def after_all(*_):
    """
    Waits for the report to be written to disk
    """
    del qat.Globals.current_report



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Functions related to touch events
"""

from qat.internal.application_context import ApplicationContext
from qat.internal import find_object
from qat.internal import mouse_operations
from qat.test_settings import Settings


def tap(
        app_context: ApplicationContext,
        definition: dict,
        mode: str,
        x=None,
        y=None,
        modifier=mouse_operations.Modifier.NONE):
    """
    Press, release, tap or move finger on the given widget using the given parameters.
    """
    definition = find_object.object_to_definition(definition)
    find_object.wait_for_object(app_context, definition)

    args = {
        'modifier': modifier
    }
    if x is not None and y is not None:
        args['x'] = x
        args['y'] = y
    command = {}
    command['command'] = 'touch'
    command['object'] = definition
    command['attribute'] = mode
    command['args'] = args

    result = app_context.send_command(command)

    if 'warning' in result:
        raise RuntimeWarning(result['warning'])


def drag(
        app_context: ApplicationContext,
        definition: dict,
        mode: str,
        x=None,
        y=None,
        dx=0,
        dy=0,
        modifier=mouse_operations.Modifier.NONE):
    """
    Drag finger on the given widget using the given parameters.
    """
    definition = find_object.object_to_definition(definition)
    find_object.wait_for_object(app_context, definition)

    args = {
        'modifier': modifier
    }
    if x is not None and y is not None:
        args['x'] = x
        args['y'] = y
    args['dx'] = dx
    args['dy'] = dy
    command = {}
    command['command'] = 'touch'
    command['object'] = definition
    command['attribute'] = mode
    command['args'] = args

    result = app_context.send_command(command, Settings.long_operation_timeout)

    if 'warning' in result:
        raise RuntimeWarning(result['warning'])



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Functions related to object identification
"""

from copy import deepcopy
import time

from qat.test_settings import Settings
from qat.internal.application_context import ApplicationContext
from qat.internal.qt_object import QtObject


def object_to_definition(object_or_def, accessible_only = False) -> dict:
    """
    Return the definition dictionary of the given object.
    If accessible_only is True, recursively append 'visible': True 
    and 'enabled': True to definition.
    """
    if isinstance(object_or_def, QtObject):
        definition = object_or_def.get_definition()
    else:
        definition = object_or_def
    for key in definition:
        if isinstance(definition[key], (QtObject,dict)):
            definition[key] = object_to_definition(definition[key], accessible_only)

    if accessible_only:
        if 'visible' not in definition:
            definition['visible'] = True
        if 'enabled' not in definition:
            definition['enabled'] = True

    return definition


def list_top_windows(
        app_context: ApplicationContext) -> list:
    """
    Return all the top windows of the application.
    """
    command = {}
    command['command'] = 'list'
    command['attribute'] = 'topWindows'

    result = app_context.send_command(command)

    objects = []
    if result['object']:
        for obj in result['object']:
            objects.append(QtObject(app_context, obj))
    return objects


def find_all_objects(
        app_context: ApplicationContext,
        definition: dict) -> list:
    """
    Return all objects matching the given definition.
    """
    definition = deepcopy(object_to_definition(definition))
    command = {}
    command['command'] = 'list'
    command['attribute'] = 'object'
    command['object'] = definition

    result = app_context.send_command(command)

    objects = []
    if result['object']:
        for obj in result['object']:
            objects.append(QtObject(app_context, obj))
    return objects


def wait_for_object_exists(
        app_context: ApplicationContext,
        definition: dict,
        timeout=Settings.wait_for_object_timeout) -> QtObject:
    """
    Wait for the given object to exist in the AUT.
    """
    start_time = round(1000 * time.time())
    definition = deepcopy(object_to_definition(definition))
    command = {}
    command['command'] = 'find'
    command['object'] = definition

    last_error = None
    while (round(1000 * time.time()) - start_time) < timeout:
        try:
            app_context.send_command(command, timeout)
            return QtObject(app_context, definition)
        except LookupError as error:
            last_error = error
            time.sleep(0.2)

    if last_error is not None:
        raise last_error
    return None


def wait_for_object(
        app_context: ApplicationContext,
        definition: dict,
        timeout=Settings.wait_for_object_timeout) -> QtObject:
    """
    Wait for the given object to be accessible (i.e visible and enabled) in the AUT.
    """
    local_definition = deepcopy(object_to_definition(definition))
    local_definition = object_to_definition(local_definition, accessible_only=True)
    return wait_for_object_exists(app_context, local_definition, timeout)


def wait_for_object_missing(
        app_context: ApplicationContext,
        definition: dict,
        timeout=Settings.wait_for_object_timeout):
    """
    Wait for the given object to be deleted from the AUT.
    """
    start_time = round(1000 * time.time())

    while (round(1000 * time.time()) - start_time) < timeout:
        try:
            result = wait_for_object_exists(app_context, definition, 100)
            if result is None or result.is_null():
                return
            time.sleep(0.2)
        except LookupError:
            return

    raise TimeoutError(f'Object "{definition}" still exists after {timeout} ms')



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Provides the ApplicationProxy class
"""

from dataclasses import dataclass
from pathlib import Path
import json
import os
import shlex
import subprocess

from qat.test_settings import Settings
# Avoid circular imports
# pylint: disable = consider-using-from-import
import qat.internal.tcp_client as tcp_client
import qat.internal.tcp_server as tcp_server

class ApplicationContext():
    """
    Class implementing a proxy to interact with a running application
    """

    @dataclass
    class Internal():
        """
        Class holding internal attributes
        """
        def __init__(self):
            self.process = None
            self.client = None
            self.server = None


    def __init__(self, name: str, path: str) -> None:
        self.name = name
        self.path = Path(path)
        self.pid = -1
        self.qt_version = None
        self.return_code = None
        self.config_file = None
        self._internal = ApplicationContext.Internal()
        self._internal.process = None
        self._internal.client = None
        self._internal.server = tcp_server.TcpServer(self)
        self._internal.server.start()


    def get_name(self) -> str:
        """
        Return the name of the running application
        """
        return self.name


    def get_path(self) -> str:
        """
        Return the parent (folder) path of the running application
        """
        return self.path.parent


    def get_app_path(self) -> str:
        """
        Return the path of the running application
        """
        return self.path


    def launch(self, args: str, local_env: dict, is_posix=True):
        """
        Launch the application with the given args and environment
        """
        command = [str(self.path)]
        if args and len(args) > 0:
            command = command + shlex.split(args, is_posix)

        # process must stay alive
        # pylint: disable = consider-using-with
        self._internal.process = subprocess.Popen(
            command,
            cwd=str(self.path.parent),
            env=local_env
        )
        self.pid = self._internal.process.pid
        print(f"Application started with PID {self.pid}")


    def init_comm(self, port: int, timeout=None):
        """
        Connect a remote C++ client to the given application context's server
        """
        if timeout is None:
            timeout = Settings.wait_for_object_timeout
        self._internal.client = tcp_client.TcpClient(port)
        host = self._internal.server.get_host()
        server_port = self._internal.server.get_port()
        command = {}
        command['command'] = 'communication'
        command['attribute'] = 'init'
        command['args'] = {'host': host, 'port': server_port}

        self.send_command(command, timeout)


    def init_version_info(self, timeout=None):
        """
        Retrieve and store version info (e.g. current Qt version)
        """
        if timeout is None:
            timeout = Settings.wait_for_object_timeout
        command = {}
        command['command'] = 'list'
        command['attribute'] = 'versionInfo'

        result = self.send_command(command, timeout)
        self.qt_version = result['qtVersion']


    def send_command(self, command: dict, timeout = None):
        """
        Send the given command to the socket and return the response as a json object
        """
        if timeout is None:
            timeout = Settings.wait_for_object_timeout
        result = self._internal.client.send_command(
            json.dumps(command),
            timeout)
        result = json.loads(result)
        if 'error' in result:
            if command['command'] in ['find', 'list']:
                raise LookupError(result['error'])

            raise RuntimeError(result['error'])
        return result


    def register_callback(self, callback_id, callback) -> int:
        """
        Register the given callback.
        """
        return self._internal.server.register_callback(callback_id, callback)


    def unregister_callback(self, callback_id):
        """
        Unregister the given callback
        """
        self._internal.server.unregister_callback(callback_id)


    def is_running(self) -> bool:
        """
        Return whether the application is running or not
        """
        if self.pid < 0 or self.return_code is not None:
            return False
        if self._internal.process is not None:
            self.return_code = self._internal.process.poll()
        if self.return_code is not None:
            return False
        if self._internal.client is not None:
            return self._internal.client.is_connected()

        return False


    def is_finished(self) -> bool:
        """
        Return whether the application has been closed or not
        """
        if self.pid <= 0:
            return False
        if self.return_code is None and self._internal.process is not None:
            self.return_code = self._internal.process.poll()
        if self.return_code is not None:
            return True
        return False


    def get_exit_code(self) -> int:
        """
        Wait for the application to close then return its exit code.
        """
        if self._internal.process:
            return self._internal.process.wait(Settings.wait_for_app_stop_timeout / 1000)
        if self.return_code is not None:
            return self.return_code

        raise ProcessLookupError(
            "Cannot retrieve exit code: process does not exist")


    def kill(self):
        """
        Kill the current application
        """
        if self.config_file is not None:
            if Path(self.config_file).is_file():
                os.remove(self.config_file)
        if self._internal.client is not None:
            self._internal.client.close()
        if self._internal.process is not None:
            self._internal.process.kill()
        elif self.pid > 0:
            os.kill(self.pid, 9)
            self.return_code = 9
        if self._internal.server is not None:
            self._internal.server.stop()



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
A simple TCP server
"""

from threading import Lock, Thread

import json
import socketserver

from qat.internal.qt_custom_object import QtCustomObject

callbacks_lock = Lock()
callbacks = {}


class QatRequestHandler(socketserver.StreamRequestHandler):
    """
    Request handler derived from StreamRequestHandler
    """
    def handle(self):
        print(f'New connection from {self.client_address[0]}')
        while True:
            try:
                header = self.rfile.readline().strip()
            except ConnectionResetError:
                break
            except: # pylint: disable=bare-except
                print('Error reading connected socket')
                break
            if header == b'':
                print('Connection was closed by the client')
                break

            try:
                length = int(header)
                content = self.rfile.read(length).decode('utf-8')
                self.execute_callback(json.loads(content))
            except Exception as error: # pylint: disable=broad-exception-caught
                # Avoid stopping the server upon errors
                print('Error in request handler: ' + str(error))


    def execute_callback(self, content: dict):
        """
        Execute the callback 
        """
        # Avoid cyclic import
        # pylint: disable = import-outside-toplevel
        # pylint: disable = cyclic-import
        from qat.internal.qt_object import QtObject
        callback_id = content['id']
        with callbacks_lock:
            if callback_id in callbacks:
                callback_elements = callbacks[callback_id]
                context = callback_elements[0]
                callback = callback_elements[1]
            else:
                print(f"Unknown callback ID: {callback_id}")
                return

        if 'args' in content:
            arg = content['args']
            if 'value' in arg:
                value = arg['value']
                if isinstance(value, dict):
                    callback(QtCustomObject(value))
                else:
                    callback(value)
            elif 'object' in arg:
                callback(QtObject(context, arg['object']))

        else:
            callback()


class TcpServer():
    """
    Class implementing a TCP server
    """
    daemon_threads = True

    def __init__(self, context, host="127.0.0.1", port=None) -> None:
        """
        Constructor.
        """
        self._context = context
        self._host = host
        self._port = port
        self._server = None
        self._thread = None

    def start(self):
        """
        Start the server thread
        """
        self._server = socketserver.ThreadingTCPServer(
            (self._host, 0), QatRequestHandler)
        self._port = self._server.socket.getsockname()[1]
        self._thread = Thread(target=self._internal_serve, daemon=True)
        self._thread.start()

    def stop(self):
        """
        Stop the server thread
        """
        if self._server is not None:
            self._server.shutdown()
            self._server.server_close()
            self._server = None
        if self._thread is not None:
            self._thread.join()

    def _internal_serve(self):
        try:
            self._server.serve_forever()
        except Exception: # pylint: disable=broad-exception-caught
            # Avoid raising exceptions from threads
            pass

    def get_host(self) -> str:
        """
        Return the host address
        """
        return self._host

    def get_port(self) -> int:
        """
        Return the server port
        """
        return self._port

    def register_callback(self, callback_id, callback) -> int:
        """
        Register the given callback.
        """
        with callbacks_lock:
            callbacks[callback_id] = (self._context, callback)

        return callback_id

    def unregister_callback(self, callback_id):
        """
        Unregister the given callback
        """
        with callbacks_lock:
            if callback_id in callbacks:
                del callbacks[callback_id]



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Functions used for development and debugging purposes
"""

from qat.internal.application_context import ApplicationContext


def activate_picker(app_context: ApplicationContext):
    """
    Activate the object picker
    """
    command = {}
    command['command'] = 'action'
    command['attribute'] = 'picker'
    command['args'] = 'enable'

    app_context.send_command(command)


def deactivate_picker(app_context: ApplicationContext):
    """
    Deactivate the object picker
    """
    command = {}
    command['command'] = 'action'
    command['attribute'] = 'picker'
    command['args'] = 'disable'

    app_context.send_command(command)


def lock_application(app_context: ApplicationContext):
    """
    Lock application GUI by filtering external/user events
    """
    command = {}
    command['command'] = 'action'
    command['attribute'] = 'lock'
    command['args'] = 'enable'

    app_context.send_command(command)


def unlock_application(app_context: ApplicationContext):
    """
    Unlock application GUI by allowing external/user events
    """
    command = {}
    command['command'] = 'action'
    command['attribute'] = 'lock'
    command['args'] = 'disable'

    app_context.send_command(command)



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Global definition of Qat constants
"""

# Global timeout in ms when waiting for objects
WAIT_FOR_OBJECT_TIMEOUT = 3 * 1000
WAIT_FOR_APP_START = 60 * 1000
WAIT_FOR_APP_STOP = 30 * 1000

# Extended timeout for long operation (e.g. drag-and-drop)
LONG_OPERATION_TIMEOUT = 10 * 1000

# Actions on failure
SCREENSHOT_ON_FAIL = True
CONTINUE_AFTER_FAIL = False

# Define when the GUI should be locked (i.e. ignoring user inputs)
# Can be one of "auto" (default), "always" or "never"
LOCK_UI = "auto"



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Generate default test suite from template
"""

from importlib import resources as res
from pathlib import Path

import argparse
import os
import shutil


def create_suite():
    """
    Create a default suite
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('suite_type', nargs='?', default='bdd', choices=['bdd', 'script'], help="'script' or 'bdd' (default)")
    parser.add_argument('--dest', required=False)
    args = parser.parse_args()
    templates_path = res.files('qat') / 'templates'
    if not templates_path.is_dir():
        templates_path = Path(os.getcwd()) / 'templates'
    destination = args.dest or os.getcwd()
    destination = Path(destination)
    os.makedirs(destination, exist_ok=True)

    common_files = ['applications.json', 'testSettings.json']
    for file in common_files:
        if not os.path.exists(destination / file):
            shutil.copyfile(templates_path / file, destination / file)
    if args.suite_type == 'bdd':
        bdd_files=[
            'environment.py',
        ]
        destination = destination / 'features/'
        os.makedirs(destination, exist_ok=True)
        for file in bdd_files:
            if not os.path.exists(destination / file):
                shutil.copyfile(templates_path / file, destination / file)
        bdd_folders=[
            'demo',
            'scripts',
            'steps'
        ]
        for folder in bdd_folders:
            os.makedirs(destination / folder, exist_ok=True)
            shutil.copytree(templates_path / folder, destination / folder, dirs_exist_ok=True)
    else:
        script_files=[
            'demo.py',
            'scripts/object_dictionary.py'
        ]
        os.makedirs(destination / 'scripts', exist_ok=True)
        for file in script_files:
            if not os.path.exists(destination / file):
                shutil.copyfile(templates_path / file, destination / file)


if __name__ == "__main__":
    create_suite()


# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Provides the ApplicationManager class
"""

from pathlib import Path
import os

import tkinter as tk
from tkinter import ttk
from tkinter import filedialog, messagebox

from qat.gui import gui_utils
from qat.gui.spy_window import SpyWindow
from qat import app_launcher
import qat

# pylint: disable=attribute-defined-outside-init

class ApplicationManager():
    """
    Show a GUI to manage applications
    """
    def __init__(self, root: tk.Tk) -> None:
        self._config_file = qat.get_config_file()
        self._image_path = Path(os.path.dirname(
            __file__)).resolve().absolute() / 'images'
        self._root = root
        self.is_editing = False
        self.is_adding = False
        self.applications = {}
        self.generate_start_window()


    def generate_start_window(self):
        """
        Create components the ApplicationManager window
        """
        self._start_window = tk.Toplevel(self._root)
        self._start_window.title('Qat Application Manager')
        self._start_window.resizable(False, False)
        self._start_window.protocol("WM_DELETE_WINDOW", self._root.destroy)

        self._generate_toolbar()
        self._generate_editing_area()


    def _generate_editing_area(self):
        image_dir = self._image_path
        # Add text field frame
        self.text_fields = tk.Frame(self._start_window, padx = 5, pady = 3)
        self.text_fields.grid(row=1, column=0, sticky='w')

        # Browse button
        self.browse_button = tk.Button(
            self.text_fields, text="Browse...", command=self.browse_application, state=tk.DISABLED)
        self.browse_button.grid(column=2, row=0)

        # Text
        self.path_label = tk.Label(self.text_fields, text="Path:")
        self.path_label.grid(row=0, column=0, sticky='w')
        self.args_label = tk.Label(self.text_fields, text="Arguments:")
        self.args_label.grid(row=2, column=0, sticky='w')

        # scrollbars
        self.path_scrollbar = ttk.Scrollbar(
            self.text_fields, orient='horizontal')
        self.path_scrollbar.grid(row=1, column=1, sticky='ew')
        self.args_scrollbar = ttk.Scrollbar(
            self.text_fields, orient='horizontal')
        self.args_scrollbar.grid(row=3, column=1, sticky='ew')

        # Text fields
        self.text_widget_path = tk.Entry(
            self.text_fields, width=100, state=tk.NORMAL,
            xscrollcommand=self.path_scrollbar.set, disabledforeground="gray30")
        if self.combobox.get() in self.applications:
            self.text_widget_path.insert(
                tk.END, self.applications[self.combobox.get()]['path'])
        self.text_widget_path["state"] = tk.DISABLED
        self.text_widget_path.grid(row=0, column=1, sticky='ew')
        self.path_scrollbar.configure(command=self.text_widget_path.xview)

        self.text_widget_args = tk.Entry(
            self.text_fields, width=100, state=tk.NORMAL,
            xscrollcommand=self.args_scrollbar.set, disabledforeground="gray30")
        if self.combobox.get() in self.applications:
            self.text_widget_args.insert(
                tk.END, self.applications[self.combobox.get()]['args'])
        self.text_widget_args["state"] = tk.DISABLED
        self.text_widget_args.grid(row=2, column=1, sticky='ew')
        self.args_scrollbar.configure(command=self.text_widget_args.xview)

        self.spy_window = SpyWindow(self._root, self._start_window)
        self.spy_window.generate_spy_window(image_dir)


    def _generate_toolbar(self):
        image_dir = self._image_path
        # Add toolbar
        toolbar = tk.Frame(self._start_window, padx=5, pady=3)
        toolbar.grid(row=0, column=0, sticky='w')

        # Combobox creation
        self.combobox = ttk.Combobox(toolbar, state="readonly", width=50)
        self.applications = qat.list_applications()
        if len(self.applications) > 0:
            self.combobox['values'] = list(i for i in self.applications)
            self.combobox.current(0)

        self.combobox.grid(row=0, column=0, padx=(0, 5), sticky='nsw')

        self.combobox.bind("<<ComboboxSelected>>", self._update_text_fields)

        # Start button
        self.start_icon = tk.PhotoImage(file=image_dir / "start_icon.png")
        self.start_button = tk.Button(
            toolbar, text="Start", image=self.start_icon, command=self.start_application)
        self.start_button.grid(column=1, row=0, sticky='w')

        # Add button
        self.add_icon = tk.PhotoImage(file=image_dir / "add_icon.png")
        self.add_button = tk.Button(
            toolbar, text="Add", image=self.add_icon, command=self.add_application)
        self.add_button.grid(column=3, row=0, sticky='w')

        # Edit button
        self.edit_icon = tk.PhotoImage(file=image_dir / "edit_icon.png")
        self.edit_button = tk.Button(
            toolbar, text="Edit", image=self.edit_icon, command=self.edit_application)
        self.edit_button.grid(column=2, row=0, sticky='w')

        # Save button
        self.save_icon = tk.PhotoImage(file=image_dir / "save_icon.png")
        self.save_button = tk.Button(
            toolbar,
            text="Save",
            image=self.save_icon,
            command=self.register_application,
            state=tk.DISABLED)
        self.save_button.grid(column=4, row=0, sticky='w')

        # Delete button
        self.delete_icon = tk.PhotoImage(file=image_dir / "delete_icon.png")
        self.delete_button = tk.Button(
            toolbar, text="Delete", image=self.delete_icon, command=self.open_delete_dialog)
        self.delete_button.grid(column=5, row=0)

        # Create a tooltip for each buttons of the toolbar
        gui_utils.create_button_tooltip(toolbar, 1.0)

    def _update_text_fields(self, _):
        if self.is_adding:
            selection = self.combobox.get()
            self.add_application()
            self.combobox.set(selection)

        self.combobox.select_clear()

        self.text_widget_path['state'] = tk.NORMAL
        self.text_widget_path.delete(0, tk.END)
        self.text_widget_path.insert(
            tk.END, self.applications[self.combobox.get()]['path'])
        self.text_widget_args['state'] = tk.NORMAL
        self.text_widget_args.delete(0, tk.END)
        self.text_widget_args.insert(
            tk.END, self.applications[self.combobox.get()]['args'])

        self.text_widget_path['state'] = tk.DISABLED
        self.text_widget_args['state'] = tk.DISABLED

    def start_application(self):
        """
        Launch the selected application
        """
        self._start_window.withdraw()

        app_name = self.combobox.get()
        try:
            qat.start_application(app_name, self.applications[app_name]['args'])
            qat.unlock_application()
        except Exception as error: # pylint: disable=broad-exception-caught
            messagebox.showerror(
                "Error",
                f"Could not launch application: {str(error)}",
                parent=self._start_window)
            self._start_window.deiconify()
            return

        self.spy_window.populate_tree([])
        self.spy_window.show()


    def edit_application(self):
        """
        Edit selected application (path and args)
        """
        if not self.is_editing:
            self.is_editing = True
            self.start_button['state'] = tk.DISABLED
            self.add_button['state'] = tk.DISABLED
            self.delete_button['state'] = tk.DISABLED
            self.combobox['state'] = tk.DISABLED
            self.save_button['state'] = tk.NORMAL
            self.browse_button['state'] = tk.NORMAL
            self.text_widget_path['state'] = tk.NORMAL
            self.text_widget_args['state'] = tk.NORMAL
        else:
            self.is_editing = False
            self.start_button['state'] = tk.NORMAL
            self.add_button['state'] = tk.NORMAL
            self.delete_button['state'] = tk.NORMAL
            self.combobox['state'] = 'readonly'
            self.save_button['state'] = tk.DISABLED
            self.browse_button['state'] = tk.DISABLED

            # discard changes
            self.text_widget_path.delete(0, tk.END)
            self.text_widget_path.insert(
                tk.END, self.applications[self.combobox.get()]['path'])
            self.text_widget_args.delete(0, tk.END)
            self.text_widget_args.insert(
                tk.END, self.applications[self.combobox.get()]['args'])

            self.text_widget_path['state'] = tk.DISABLED
            self.text_widget_args['state'] = tk.DISABLED
        self.edit_button.config(
            relief="sunken" if self.is_editing else "raised")


    def add_application(self):
        """
        Define a new application
        """
        self.is_adding = not self.is_adding
        self.add_button.config(relief="sunken" if self.is_adding else "raised")
        if self.is_adding:
            self.start_button['state'] = tk.DISABLED
            self.edit_button['state'] = tk.DISABLED
            self.delete_button['state'] = tk.DISABLED
            self.save_button['state'] = tk.NORMAL
            self.browse_button['state'] = tk.NORMAL

            self.combobox['state'] = tk.NORMAL
            self.combobox.set('')

            self.text_widget_path['state'] = tk.NORMAL
            self.text_widget_path.delete(0, tk.END)

            self.text_widget_args['state'] = tk.NORMAL
            self.text_widget_args.delete(0, tk.END)
        else:
            self.start_button['state'] = tk.NORMAL
            self.edit_button['state'] = tk.NORMAL
            self.delete_button['state'] = tk.NORMAL
            self.save_button['state'] = tk.DISABLED
            self.browse_button['state'] = tk.DISABLED

            self.combobox['state'] = 'readonly'
            self.combobox.current(0)

            self.text_widget_path.delete(0, tk.END)
            self.text_widget_args.delete(0, tk.END)
            self.text_widget_path.insert(
                tk.END, self.applications[self.combobox.get()]['path'])
            self.text_widget_args.insert(
                tk.END, self.applications[self.combobox.get()]['args'])
            self.text_widget_path['state'] = tk.DISABLED
            self.text_widget_args['state'] = tk.DISABLED


    def register_application(self):
        """
        Register a new application to the configuration file
        """

        if len(self.combobox.get()) == 0 or len(self.text_widget_path.get().strip()) == 0:
            return

        qat.register_application(
            self.combobox.get(),
            self.text_widget_path.get().strip(),
            self.text_widget_args.get().strip(),
            None)

        self.combobox['state'] = 'readonly'
        self.text_widget_path['state'] = tk.DISABLED
        self.text_widget_args['state'] = tk.DISABLED
        self.applications = qat.list_applications()
        self.combobox['values'] = list(i for i in self.applications)

        self.start_button['state'] = tk.NORMAL
        self.edit_button['state'] = tk.NORMAL
        self.add_button['state'] = tk.NORMAL
        self.delete_button['state'] = tk.NORMAL
        self.save_button['state'] = tk.DISABLED
        self.browse_button['state'] = tk.DISABLED
        self.is_editing = False
        self.edit_button.config(relief="raised")


    def browse_application(self):
        """
        Use a File selector dialog to select an application
        """
        extension = '*'
        if app_launcher.is_windows():
            extension = '*.exe'
        filename = filedialog.askopenfilename(
            initialdir="C:/",
            title="Select a File",
            filetypes=[("Applications", extension)],
            parent=self._start_window
        )

        if filename:
            self.text_widget_path.delete(0, tk.END)
            self.text_widget_path.insert(tk.END, filename)


    def open_delete_dialog(self):
        """
        Display a confirmation dialog when deleting an application
        """
        answer = messagebox.askokcancel(
            "Remove application",
            "Are you sure you want to delete '" +
            self.combobox.get() + "' from the application list?",
            parent=self._start_window)
        if answer:
            self.delete_application()


    def delete_application(self):
        """
        Remove an application for the configuration file
        """
        qat.unregister_application(self.combobox.get(), None)

        self.combobox['state'] = 'readonly'
        self.text_widget_path['state'] = tk.NORMAL
        self.text_widget_args['state'] = tk.NORMAL
        self.text_widget_path.delete(0, tk.END)
        self.text_widget_args.delete(0, tk.END)

        self.applications = qat.list_applications()
        self.combobox['values'] = list(i for i in self.applications)

        self.combobox.current(0)
        self.text_widget_path.insert(
            tk.END, self.applications[self.combobox.get()]['path'])
        self.text_widget_args.insert(
            tk.END, self.applications[self.combobox.get()]['args'])

        self.text_widget_path['state'] = tk.DISABLED
        self.text_widget_args['state'] = tk.DISABLED



# (c) Copyright 2023, Qat’s Authors

import os
import setuptools

qat_version = os.environ.get('QAT_VERSION')

long_description = """
# Qat (Qt Application Tester)

Qat is a testing framework for Qt-based applications.

Qat provides a Python API to interact with any existing Qt application by accessing QML/QtQuick/QWidget elements and simulating user manipulations.

It is also integrated to [behave](https://github.com/behave/behave) to support Behavior-Driven Development (BDD) with the [Gherkin language](https://cucumber.io/docs/gherkin/).

Although Qat uses the GUI to interact with the tested application, it is oriented towards BDD and functional testing rather than pure UI or non-regression testing.

The main objective of Qat is to provide quick feedback to developers and easy integration to build systems.

The complete documentation is available on [readthedocs](https://qat.readthedocs.io/en/latest/) and on Qat's [Gitlab project](https://gitlab.com/testing-tool/qat/-/blob/develop/README.md?ref_type=heads).
"""


setuptools.setup(
    name="qat",
    author="Quentin Derouault",
    author_email="qat.authors@gmail.com",
    url="https://gitlab.com/testing-tool/qat",
    version=qat_version,
    description="Qt Application Tester, a BDD-oriented framework to test Qt-based applications",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=[
        "qat",
        "qat.internal",
        "qat.bin",
        "qat.bin.plugins", 
        "qat.gui", 
        "qat.gui.images", 
        "qat.templates", 
        "qat.templates.demo", 
        "qat.templates.scripts", 
        "qat.templates.steps"],
    package_dir={
        '': 'client'
    },
    package_data={
        'qat.bin': ['*', 'plugins/*'],
        'qat.gui': ['images/*'],
        'qat.templates': ['*', '*/*']
    },
    include_package_data=True,
    exclude_package_data={'qat.bin': ['plugins']},
    entry_points={
        'console_scripts': [
            'qat-gui = qat.gui.launcher:open_gui',
            'qat-create-suite = qat.templates.generator:create_suite',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Testing :: BDD"
    ],
    python_requires='>=3.9',
    install_requires=[
        'behave',
        'sv_ttk',
        'tkinter-tooltip',
        'xmlschema'
    ],
    project_urls={
        "Documentation": f"https://qat.readthedocs.io/en/{qat_version}/"
    }
)



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
A simple TCP client
"""

from threading import Lock

import errno
import socket
import time

from qat.test_settings import Settings


class TcpClient():
    """
    Class sending commands to a TCP server
    """

    def __init__(self, port: int) -> None:
        """
        Constructor.
        Create and connect a TCP socket to the given port on localhost
        """
        self._port = port
        self._lock = Lock()
        self._connection_errors = 0
        self.connect()


    def connect(self):
        """
        Create and connect a TCP socket to the current port on localhost
        """
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect(("localhost", self._port))
            self._socket.settimeout(Settings.wait_for_app_start_timeout / 1000)
            self._fsocket = self._socket.makefile(mode="rwb")
        except socket.error as error:
            self._socket = None
            print(f'Unable to connect to server: {error}')
            if error.errno != errno.ECONNREFUSED:
                raise
            time.sleep(0.5)


    def is_connected(self):
        """
        Return whether this client is connected or not
        """
        return self._socket is not None


    def close(self):
        """
        Close the underlying socket
        """
        with self._lock:
            if self._socket:
                self._socket.close()


    def __del__(self) -> None:
        """
        Destructor.
        Close the TCP socket
        """
        self.close()


    def send_command(self, message: str, timeout=None) -> str:
        """
        Send the given message to the socket and return the response
        """
        if timeout is None:
            timeout = Settings.wait_for_object_timeout
        with self._lock:
            try:
                self._socket.settimeout(timeout / 1000)
                header = f"{len(message)}\r\n"
                self._fsocket.write(header.encode('utf-8'))
                self._fsocket.write(message.encode('utf-8'))
                self._fsocket.flush()

                header = self._fsocket.readline()
                length = int(header)
                result = self._fsocket.read(length).decode('utf-8')
                self._connection_errors = 0
                return result
            except (OSError, ValueError, AttributeError) as error:
                self._connection_errors += 1
                # Try to reconnect upon error
                print("Error sending command - trying to reconnect: " + str(error))
        if self._connection_errors > 5:
            self._connection_errors = 0
            raise ConnectionAbortedError('Lost communication with application')
        self.connect()
        return self.send_command(message, timeout)



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Functions related to screenshots
"""

from pathlib import Path
import tempfile
import time
import os

from qat.internal.application_context import ApplicationContext
from qat.internal.qt_object import QtObject
from qat.internal import find_object
from qat.test_settings import Settings


def take_screenshot(
        app_context: ApplicationContext,
        path=None):
    """
    Take a screenshot of each current main window and save 
    them to the given path. If no path is provided, screenshots will be
    saved to the 'screenshots' subfolder of the current directory
    """
    if path is None:
        path = os.getcwd() + "/screenshots/"
    path = Path(path)
    if not path.is_absolute():
        path = path.absolute()

    command = {}
    command['command'] = 'action'
    command['attribute'] = 'screenshot'
    command['args'] = str(path)
    if path.stem == path.name:
        command['args'] += '/'

    app_context.send_command(command)


def grab_screenshot(
        app_context: ApplicationContext,
        definition,
        delay=0,
        timeout=Settings.wait_for_object_timeout):
    """
    Take a screenshot of the given widget after an optional delay in ms
    """
    definition = find_object.object_to_definition(definition)
    if delay > 0:
        time.sleep(delay / 1000)
    command = {}
    command['command'] = 'action'
    command['attribute'] = 'grab'
    command['object'] = definition
    command['args'] = ''

    result = app_context.send_command(command, timeout)

    if 'found' in result and not result['found']:
        raise LookupError("Cannot grab image: object not found")
    if 'cache_uid' not in result:
        raise RuntimeError("Cannot grab image: failed to create image")

    result = {
        'cache_uid': result['cache_uid']
    }
    image = QtObject(app_context, result)
    start_time = round(1000 * time.time())
    while image.width < 0 and (round(1000 * time.time()) - start_time) < timeout:
        time.sleep(0.2)

    if image.width < 0:
        raise RuntimeError('Failed to grab image')

    with tempfile.NamedTemporaryFile(suffix='.bmp', delete=True) as temp_file:
        temp_file.close()
        image.save(temp_file.name)

        result['_temp_image_file'] = temp_file.name
        image = QtObject(app_context, result)

        return image



# -*- coding: utf-8 -*-
# (c) Copyright 2023, Qat’s Authors

"""
Test report module
"""

from pathlib import Path
import time

from qat.test_settings import Settings
from qat import Globals
import qat

TIME_FORMAT = '%Y-%m-%dT%H_%M_%S'

class PropagatedException(Exception):
    """
    Custom exception class wrapping another exception.
    Intended to be thrown from the report when fail() is called.
    """

def start_report(
        name: str,
        report_filename: Path,
        report_class=qat.XmlReport):
    """ 
    Initialize a report with writer and return it
    """
    Globals.current_report = report_class(name, report_filename)
    return Globals.current_report


def stop_report():
    """ 
    Finalize a report and stop its writer
    """
    if Globals.current_report is not None:
        Globals.current_report.end_report()


def log(text: str, log_type="LOG"):
    """
    Add a message to the report
    """
    Globals.current_report.log(text, log_type)


def verify(
        condition,
        description,
        details=""):
    """ 
    Add a verification to the report.
    If condition is False, raise an exception.
    """
    if condition:
        passed(description, details)
        return

    failed(description, details)


def passed(description, details=""):
    """ 
    Add a passed verification to the report".
    """
    Globals.current_report.passed(description, details)


def failed(description, details=""):
    """ 
    Add a failed verification to the report".
    """
    if Settings.screenshot_on_fail:
        timestamp = time.strftime(TIME_FORMAT, time.localtime())
        report_folder = Path(qat.Globals.current_report.get_folder())
        file_name = 'fail_' + timestamp + '.png'
        file_name = report_folder / 'screenshots' / 'failures' / file_name
        try:
            qat.take_screenshot(file_name)
            Globals.current_report.log(
                "Failure screenshot saved to " + file_name.as_posix())
        except: # pylint: disable = bare-except
            pass

    Globals.current_report.failed(description, details)

    raise PropagatedException(details)


def attach_image(image, name, description) -> Path:
    """
    Add the given Image to the report and return its file name
    """
    timestamp = time.strftime(TIME_FORMAT, time.localtime())
    file_name = name + '_' + timestamp + '.png'
    file_name = Globals.current_report.get_folder() / 'screenshots' / 'attached' / file_name
    image.save(file_name)
    Globals.current_report.log(f'Attached image "{name}" has been saved to "{file_name}"')
    Globals.current_report.log(description)
    return file_name
